from __future__ import annotations

import dataclasses

import gurobipy as gp
import numpy as np

import stltspref.main_constraints as main_constraints
from .main_constraints import MilpVariables

from .stl import StlFormula
from .system_model import SystemModel
from .trace import Trace


@dataclasses.dataclass(frozen=True)
class StlMilpProblemConfig:
    N: int
    delta: float
    gamma_0: float
    gamma_N: float
    gamma_unit_length: float
    use_binary_expansion: bool
    n_gamma_digits: int | None


def create_stl_milp_problem(
    milp: gp.Model,
    N: int = 2,
    delta: float = 0.01,
    gamma_0: float = 0.0,
    gamma_N: float = 1.0,
    gamma_unit_length: float = 0.01,
    use_binary_expansion: bool = False,
    n_gamma_digits: int | None = None,
) -> StlMilpProblem:
    config = StlMilpProblemConfig(
        N,
        delta,
        gamma_0,
        gamma_N,
        gamma_unit_length,
        use_binary_expansion,
        n_gamma_digits,
    )
    return StlMilpProblem(milp, config)


class StlMilpProblem:
    milp: gp.Model
    config: StlMilpProblemConfig
    milp_variables: MilpVariables
    system_model: SystemModel
    stl_spec: StlFormula

    def __init__(
        self,
        model: gp.Model,
        config: StlMilpProblemConfig,
    ):
        self.milp = model
        self.config = config

    def initialize_milp_formulation(
        self,
        spec: StlFormula,
        ignore_system_model: bool = False,
        system_states: dict[str, gp.MVar] | None = None,
    ) -> MilpVariables:
        self.stl_spec = spec

        self.milp_variables = v = main_constraints.make_milp_variables(
            self.milp,
            spec=spec,
            N=self.config.N,
            gamma_0=self.config.gamma_0,
            gamma_N=self.config.gamma_N,
            gamma_unit_length=self.config.gamma_unit_length,
        )

        if not ignore_system_model:
            for i in range(1, self.config.N + 1):
                self.system_model.set_gamma_length(i, v.gamma[i] - v.gamma[i - 1])
            system_states = self.system_model.state
        assert system_states is not None

        main_constraints.add_timeline_constraints(
            self.milp,
            v,
            self.config.gamma_0,
            self.config.gamma_N,
            self.config.gamma_unit_length,
        )
        main_constraints.add_full_stl_constraints(self.milp, v, spec)
        main_constraints.add_predicate_constraints(
            self.milp, v, system_states, self.config.delta
        )
        return v

    @property
    def delta(self) -> float:
        return self.config.delta

    @property
    def N(self) -> int:
        return self.config.N

    @property
    def state(self) -> dict[str, gp.MVar]:
        return self.system_model.state

    def search_satisfaction(self, nbsols=1, mode=0) -> bool:
        self.milp.Params.PoolSolutions = nbsols
        self.milp.Params.PoolSearchMode = mode
        # self.milp.Params.PoolGapAbs = 
        self.milp.optimize()
        return self.milp.SolCount > 0

    @property
    def has_solution(self) -> bool:
        return self.milp.SolCount > 0

    def get_gamma_result(self) -> np.ndarray:
        return np.array(self.milp_variables.get_values()['gamma'])

    def get_trace_result(self, interpolation=False) -> list[Trace]:
        traces = []
        if interpolation:
            for i in range(self.milp.SolCount):
                self.milp.Params.SolutionNumber = i
                traces.append(self.system_model.get_trace(0.1))
            return traces
        else:
            for i in range(self.milp.SolCount):
                self.milp.Params.SolutionNumber = i
                traces.append(self.system_model.get_trace())
            return traces

    def create_system_model(self) -> SystemModel:
        if not self.config.use_binary_expansion:
            self.system_model = SystemModel(self.milp, self.N)
        else:
            if self.config.n_gamma_digits is None:
                # By default, we demand that the largest possible gamma covers the double of the horizon.
                length_in_units = (
                    self.config.gamma_N - self.config.gamma_0
                ) / self.config.gamma_unit_length
                n_gamma_digits = 1 + int(
                    np.ceil(np.log2(length_in_units / self.config.N))
                )
            else:
                n_gamma_digits = self.config.n_gamma_digits
            self.system_model = SystemModel(
                self.milp, self.N, n_gamma_digits, self.config.gamma_unit_length
            )
        return self.system_model



from __future__ import annotations

import dataclasses
from calendar import c
from typing import overload

import gurobipy as gp
import numpy as np

LE = gp.GRB.LESS_EQUAL
GE = gp.GRB.GREATER_EQUAL


@dataclasses.dataclass
class LinearExpression:
    variables: list[str]
    coefficients: list[float]
    constant: float = 0.0

    def __ge__(self, other: float) -> LinearInequality:
        x = dataclasses.replace(self, constant=self.constant - other)
        return LinearInequality(x, GE, name=f"{str(self).replace(' ', '')}_>=_{other}")

    def __le__(self, other: float) -> LinearInequality:
        x = dataclasses.replace(self, constant=self.constant - other)
        return LinearInequality(x, LE, name=f"{str(self).replace(' ', '')}_<=_{other}")

    def __repr__(self) -> str:
        combined = ' + '.join(
            f"{coeff} * {var}" for var, coeff in zip(self.variables, self.coefficients)
        )
        return f"{combined} + {self.constant}"

    @classmethod
    def f(cls, *args: str | float) -> LinearExpression:
        """Create a linear expression from variables and coefficients. No constant term is allowed.

        Example:
        >>> LinearExpression.f(1.0, 'x', 2.0, 'y')
        1.0 * x + 2.0 * y + 0.0
        """
        assert len(args) % 2 == 0
        coefficients = args[::2]
        variables = args[1::2]
        return cls(list(variables), list(coefficients))  # type: ignore

    @classmethod
    def unit(cls, var: str) -> LinearExpression:
        """Create a linear expression from a single variable with coefficient 1.0.

        Example:
        >>> LinearExpression.unit('x')
        1.0 * x + 0.0
        """
        return cls([var], [1.0])

    @overload
    def eval(self, state: dict[str, np.ndarray]) -> np.ndarray: ...

    @overload
    def eval(self, state: dict[str, float]) -> float: ...

    def eval(self, state: dict) -> np.ndarray | float:
        target_states = [state[var] for var in self.variables]
        if isinstance(target_states[0], np.ndarray):
            value = (
                np.vstack(target_states).T @ np.array(self.coefficients) + self.constant
            )
        elif isinstance(target_states[0], float):
            value = (
                np.array(target_states) @ np.array(self.coefficients) + self.constant
            )
        else:
            raise ValueError(f"Unknown type of state {target_states[0]}")
        return value

    def apply_to_variables(self, state: dict[str, gp.MVar]) -> gp.MLinExpr:
        return (
            gp.vstack([state[var] for var in self.variables]).T
            @ np.array(self.coefficients)
            + self.constant
        )

    @property
    def norm(self) -> float:
        return np.linalg.norm(self.coefficients)

    def __neg__(self) -> LinearExpression:
        return dataclasses.replace(
            self, coefficients=[-c for c in self.coefficients], constant=-self.constant
        )

    def __add__(self, other: LinearExpression | float) -> LinearExpression:
        if isinstance(other, float):
            return dataclasses.replace(self, constant=self.constant + other)
        elif isinstance(other, LinearExpression) and self.variables == other.variables:
            return dataclasses.replace(
                self,
                coefficients=[
                    c1 + c2 for c1, c2 in zip(self.coefficients, other.coefficients)
                ],
                constant=self.constant + other.constant,
            )
        elif isinstance(other, LinearExpression) and set(self.variables).isdisjoint(
            other.variables
        ):
            return LinearExpression(
                self.variables + other.variables,
                self.coefficients + other.coefficients,
                self.constant + other.constant,
            )
        else:
            return NotImplemented

    def __mul__(self, other: float) -> LinearExpression:
        return dataclasses.replace(
            self,
            coefficients=[c * other for c in self.coefficients],
            constant=self.constant * other,
        )

    def __rmul__(self, other: float) -> LinearExpression:
        return dataclasses.replace(
            self,
            coefficients=[c * other for c in self.coefficients],
            constant=self.constant * other,
        )


@dataclasses.dataclass
class LinearInequality:
    expr: LinearExpression
    sense: str
    name: str

    @property
    def sign(self) -> int:
        sign = 1 if self.sense == GE else -1
        return sign

    def invert(self) -> LinearInequality:
        return dataclasses.replace(
            self, sense=LE if self.sense == GE else GE, name=f"~{self.name}"
        )

    def to_gt_zero(self) -> LinearInequality:
        if self.sense == GE:
            return self
        else:
            return LinearInequality(-self.expr, GE, name=f"sf({self.name})")

    def __matmul__(self, name: str) -> LinearInequality:
        return dataclasses.replace(self, name=name)



from __future__ import annotations

import gurobipy as gp
import numpy as np

from .constants import EPSILON, M


def if_all_then(
    conditions: list[gp.LinExpr | gp.Var], then: gp.LinExpr | gp.Var
) -> gp.TempLConstr:
    """Return a constraint that asserts if all conditions are true, then the expression is true."""
    return at_least_one(*(1 - cond for cond in conditions), then)


def at_least_one(*args: gp.LinExpr | gp.Var) -> gp.TempLConstr:
    """Return a constraint that asserts at least one of the arguments is 1."""
    if len(args) <= 1:
        raise ValueError("At least two arguments required")
    else:
        return gp.quicksum(args) >= 1  # type: ignore


def add_disjoint_enforcer(milp: gp.Model, z: gp.Var, a: tuple, b: tuple) -> None:
    """Add constraint enforcing that the intervals a and b are disjoint.
    Intervals are considered closed.
    """
    z1 = milp.addVar(vtype=gp.GRB.BINARY)
    z2 = milp.addVar(vtype=gp.GRB.BINARY)
    milp.addConstr(z1 + z2 >= 1)
    milp.addConstr(
        a[0] - b[1] >= EPSILON - M * (2 - z - z1),
    )
    milp.addConstr(
        b[0] - a[1] >= EPSILON - M * (2 - z - z2),
    )


def multiply_with_binary_variable(
    milp: gp.Model, x: gp.LinExpr | gp.Var, z: gp.Var
) -> gp.Var:
    """
    Return a variable that represents the product of continuous variable and binary variable.
    """
    y = milp.addVar(vtype=gp.GRB.CONTINUOUS, lb=-gp.GRB.INFINITY, ub=gp.GRB.INFINITY)
    milp.addConstr((z == 0) >> (y == 0))
    milp.addConstr((z == 1) >> (y == x))
    return y


def add_and_aux_var(milp: gp.Model, z1: gp.Var, z2: gp.Var, name: str) -> gp.Var:
    """Add a binary variable that represents the conjunction of two binary variables."""
    z = milp.addVar(vtype=gp.GRB.BINARY, name=name)
    milp.addConstr(z == gp.and_(z1, z2), name=f'and_{name}')
    return z


def add_or_aux_var(milp: gp.Model, z1: gp.Var, z2: gp.Var, name: str) -> gp.Var:
    """Add a binary variable that represents the disjunction of two binary variables."""
    z = milp.addVar(vtype=gp.GRB.BINARY, name=name)
    milp.addConstr(z == gp.or_(z1, z2), name=f'or_{name}')
    return z


class BinaryExpressionValue:
    def __init__(self, milp: gp.Model, name: str, size: int, unit: float = 1.0) -> None:
        self.binary = milp.addMVar(size, vtype=gp.GRB.BINARY, name=name)
        self.powers: np.ndarray = unit * np.geomspace(1, 2 ** (size - 1), num=size)

    @property
    def value(self) -> gp.LinExpr:
        return (self.binary @ self.powers).item()

    @property
    def unit(self) -> float:
        return self.powers[0].item()

    def multiply(self, milp: gp.Model, expr: gp.LinExpr | gp.Var) -> gp.LinExpr:
        return gp.LinExpr(
            self.powers.tolist(),
            [
                multiply_with_binary_variable(milp, expr, z)
                for z in self.binary.tolist()
            ],
        )


def ensure_monotonicity_by_derivative(
    milp: gp.Model, derivative: list[gp.LinExpr | gp.Var], N: int
):
    # (To ensure monotonicity in interpolation)
    # All adjacent dx must have same sign (non-negative or non-positive)
    # z = 1 then non-negative, z = 0 then non-positive
    z = milp.addVars(N, vtype=gp.GRB.BINARY)
    for i in range(N):
        milp.addConstr(
            (z[i] == 1) >> (derivative[i] >= 0),
        )
        milp.addConstr(
            (z[i] == 1) >> (derivative[i + 1] >= 0),
        )
        milp.addConstr(
            (z[i] == 0) >> (derivative[i] <= 0),
        )
        milp.addConstr(
            (z[i] == 0) >> (derivative[i + 1] <= 0),
        )



from __future__ import annotations

import gurobipy as gp
import numpy as np

from .stl_atomic import Atomic
from .stl_base import StlFormula
from .trace import Trace


class Top(StlFormula):
    def __init__(self):
        super().__init__("True")

    def negation(self) -> StlFormula:
        return Bottom()

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        return np.ones(len(trace.time)) * np.inf


class Bottom(StlFormula):
    def __init__(self):
        super().__init__("False")

    def negation(self) -> StlFormula:
        return Top()

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        return -np.ones(len(trace.time)) * np.inf


class And(StlFormula):
    def __init__(self, *phis: StlFormula):
        super().__init__(f"({'&'.join(phi.name for phi in phis)})", children=list(phis))

    def negation(self) -> StlFormula:
        return Or(*[phi.negation() for phi in self.children])

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        if memo is not None and self.name in memo:
            return memo[self.name]

        robs = [child.robust_semantics(trace, memo) for child in self.children]
        return np.minimum.reduce(robs)


class Or(StlFormula):
    def __init__(self, *phis: StlFormula):
        super().__init__(f"({'|'.join(phi.name for phi in phis)})", children=list(phis))

    def negation(self) -> StlFormula:
        return And(*[phi.negation() for phi in self.children])

    def robust_semantics(self, trace: Trace, memo: None) -> np.ndarray:
        if memo is not None and self.name in memo:
            return memo[self.name]
        robs = [child.robust_semantics(trace, memo) for child in self.children]
        return np.maximum.reduce(robs)


class Until(StlFormula):
    def __init__(self, phi1: StlFormula, phi2: StlFormula):
        super().__init__(f"({phi1.name})Until({phi2.name})", children=[phi1, phi2])

    @property
    def phi1(self) -> StlFormula:
        return self.children[0]

    @property
    def phi2(self) -> StlFormula:
        return self.children[1]

    def negation(self) -> StlFormula:
        return Release(self.phi1.negation(), self.phi2.negation())

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        if memo is not None and self.name in memo:
            return memo[self.name]

        rho1 = self.phi1.robust_semantics(trace, memo)
        rho2 = self.phi2.robust_semantics(trace, memo)
        return np.array([
            np.max([
                np.minimum(rho2[j], rho1[i : j + 1].min()) for j in range(i, len(rho1))
            ])
            for i in range(len(rho1))
        ])


class Release(StlFormula):
    def __init__(self, phi1: StlFormula, phi2: StlFormula):
        super().__init__(f"({phi1.name})Release({phi2.name})", children=[phi1, phi2])

    @property
    def phi1(self) -> StlFormula:
        return self.children[0]

    @property
    def phi2(self) -> StlFormula:
        return self.children[1]

    def negation(self) -> StlFormula:
        return Until(self.phi1.negation(), self.phi2.negation())

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        if memo is not None and self.name in memo:
            return memo[self.name]

        rho1 = self.phi1.robust_semantics(trace, memo)
        rho2 = self.phi2.robust_semantics(trace, memo)
        return np.array([
            np.min([
                np.maximum(rho2[j], rho1[i : j + 1].max()) for j in range(i, len(rho1))
            ])
            for i in range(len(rho1))
        ])


class Ev(StlFormula):
    def __init__(self, phi: StlFormula):
        super().__init__(f"Ev({phi.name})", children=[phi])

    def negation(self) -> StlFormula:
        return Alw(self.child.negation())

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        if memo is not None and self.name in memo:
            return memo[self.name]

        rho = self.child.robust_semantics(trace, memo)
        return np.array([np.max(rho[i:]) for i in range(len(rho))])


class Alw(StlFormula):
    def __init__(self, phi: StlFormula):
        super().__init__(f"Alw({phi.name})", children=[phi])

    def negation(self) -> StlFormula:
        return Ev(self.child.negation())

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        if memo is not None and self.name in memo:
            return memo[self.name]

        rho = self.child.robust_semantics(trace, memo)
        return np.array([np.min(rho[i:]) for i in range(len(rho))])


class BoundedAlw(StlFormula):
    interval: tuple[float, float]

    def __init__(self, interval: tuple[float, float], phi: StlFormula):
        super().__init__(
            f"Alw_[{interval[0]},{interval[1]}]({phi.name})", children=[phi]
        )
        self.bounded = True
        self.interval = interval

    def negation(self) -> StlFormula:
        return BoundedEv(self.interval, self.child.negation())

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        if memo is not None and self.name in memo:
            return memo[self.name]

        rho = self.child.robust_semantics(trace, memo)

        interval = self.interval
        if isinstance(interval[0], gp.Var):
            interval = (interval[0].X, interval[1])
        if isinstance(interval[1], gp.Var):
            interval = (interval[0], interval[1].X)

        def intersecting_index(i: int):
            if trace.time[i] + interval[0] > trace.time[-1]:
                return trace.time == trace.time[-1]
            return (trace.time >= trace.time[i] + interval[0]) & (
                trace.time <= trace.time[i] + interval[1]
            )

        return np.array([rho[intersecting_index(i)].min() for i in range(len(rho))])


class BoundedEv(StlFormula):
    interval: tuple[float, float]

    def __init__(self, interval: tuple[float, float], phi: StlFormula):
        super().__init__(
            f"Ev_[{interval[0]},{interval[1]}]({phi.name})", children=[phi]
        )
        self.bounded = True
        self.interval = interval

    def negation(self) -> StlFormula:
        return BoundedAlw(self.interval, self.child.negation())

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        if memo is not None and self.name in memo:
            return memo[self.name]

        rho = self.child.robust_semantics(trace, memo)

        interval = self.interval
        if isinstance(interval[0], gp.Var):
            interval = (interval[0].X, interval[1])
        if isinstance(interval[1], gp.Var):
            interval = (interval[0], interval[1].X)

        def intersecting_index(i: int):
            if trace.time[i] + interval[0] > trace.time[-1]:
                return trace.time == trace.time[-1]
            return (trace.time >= trace.time[i] + interval[0]) & (
                trace.time <= trace.time[i] + interval[1]
            )

        return np.array([rho[intersecting_index(i)].max() for i in range(len(rho))])


def BoundedUntil(
    interval: tuple[float, float], phi1: StlFormula, phi2: StlFormula
) -> StlFormula:
    if interval[0] == 0:
        until = Until(phi1, phi2)
    else:
        until = BoundedAlw((0, interval[0]), Until(phi1, phi2))
    return And(BoundedEv(interval, phi2), until)


def BoundedRelease(
    interval: tuple[float, float], phi1: StlFormula, phi2: StlFormula
) -> StlFormula:
    if interval[0] == 0:
        release = Release(phi1, phi2)
    else:
        release = BoundedEv((0, interval[0]), Release(phi1, phi2))
    return Or(BoundedAlw(interval, phi2), release)


def Implies(phi1: StlFormula, phi2: StlFormula) -> StlFormula:
    return Or(phi1.negation(), phi2)


def atomic_predicates(phi: StlFormula) -> list[Atomic]:
    """Return a list of atomic predicates in the formula."""
    if isinstance(phi, Atomic):
        return [phi]
    elif isinstance(phi, BoundedEv) and isinstance(phi.phantom_child, Atomic):
        return [phi.phantom_child]
    else:
        result = []
        for child in phi.children:
            candidates = atomic_predicates(child)
            for new in candidates:
                if new.name not in [p.name for p in result]:
                    result.append(new)
        return result


def make_unique(root: StlFormula):
    formula_visited = {}

    for phi in root.get_subformulas():
        if phi.name not in formula_visited:
            formula_visited[phi.name] = phi
        for i, child in enumerate(phi.children):
            if child.name in formula_visited:
                phi.children[i] = formula_visited[child.name]

    return formula_visited[root.name]



import numpy as np
import pandas as pd
import plotly.express as px

from stltspref.trace import Trace


def plot_rss(trace_df: pd.DataFrame, interpolate: bool = False):
    from stltspref.rss_scenario import car_length, car_width

    if interpolate:
        trace = Trace.from_df(trace_df).interpolate().df
    else:
        trace = trace_df.copy()

    trace['distance_x'] = abs(trace['bx'] - trace['ax'])
    trace['distance_y'] = abs(trace['by'] - trace['ay'])
    trace['distance'] = np.maximum(
        abs(trace['by'] - trace['ay']) - car_width,
        abs(trace['bx'] - trace['ax']) - car_length,
    )

    trace['danger'] = -np.maximum.reduce([
        trace['bx'] - trace['ax'] - trace['rssDistance_a'] - car_length,
        trace['ax'] - trace['bx'] - trace['rssDistance_b'] - car_length,
        trace['by'] - trace['ay'] - trace['lateralRssDistance_a'] - car_width,
        trace['ay'] - trace['by'] - trace['lateralRssDistance_b'] - car_width,
    ])
    trace_a = pd.DataFrame(
        dict(
            t=trace['t'],
            x=trace['ax'],
            y=trace['ay'],
            x_vel=trace['ax_vel'],
            x_acc=trace['ax_acc'],
            y_vel=trace['ay_vel'],
            y_acc=trace['ay_acc'],
            car='a',
        )
    )

    trace_b = pd.DataFrame(
        dict(
            t=trace['t'],
            x=trace['bx'],
            y=trace['by'],
            x_vel=trace['bx_vel'],
            x_acc=trace['bx_acc'],
            y_vel=trace['by_vel'],
            y_acc=trace['by_acc'],
            car='b',
        )
    )

    px.line(
        trace,
        x='t',
        y=[
            'distance',
            'rssDistance_a',
            'rssDistance_b',
            'lateralRssDistance_a',
            'lateralRssDistance_b',
            'danger',
        ],
        markers=not interpolate,
    ).show()

    px.line(
        trace,
        x='t',
        y=[
            'ax',
            'ay',
            'ax_vel',
            'ax_acc',
            'ay_vel',
            'ay_acc',
            'bx',
            'by',
            'bx_vel',
            'bx_acc',
            'by_vel',
            'by_acc',
        ],
        markers=not interpolate,
    ).show()

    fig = px.line(
        pd.concat(
            [trace_a, trace_b],
            ignore_index=True,
        ),
        x='x',
        y='y',
        color='car',
        hover_data=['t'],
        markers=not interpolate,
    )

    for time in [0, trace['danger'].argmax(), len(trace) - 1]:
        fig.add_shape(
            type="rect",
            x0=trace['ax'][time],
            y0=trace['ay'][time],
            x1=trace['ax'][time] + car_length,
            y1=trace['ay'][time] - car_width,
            fillcolor="LightSkyBlue",
            line=dict(color="RoyalBlue"),
        )
        fig.add_shape(
            type="rect",
            x0=trace['bx'][time],
            y0=trace['by'][time],
            x1=trace['bx'][time] + car_length,
            y1=trace['by'][time] - car_width,
            fillcolor="LightSalmon",
            line=dict(color="Tomato"),
        )
    for l in [0, 3.5, 7, 10.5]:
        fig.add_hline(y=l, line_width=1, line_dash="dash", line_color="gray")
    fig.show()



from __future__ import annotations

import gurobipy as gp
import numpy as np
from scipy import integrate

from .linear_expression import LinearExpression as L
from .linear_expression import LinearInequality
from .milp_helper import BinaryExpressionValue
from .trace import Trace
from .validator import ValidatorRules


class SystemModel:
    """
    The system model.
    """

    def __init__(
        self,
        model: gp.Model,
        N: int,
        n_gamma_digits: int | None = None,
        gamma_unit_length: float = 0.01,
    ):
        self.milp = model
        self.N = N
        self.state: dict[str, gp.MVar] = {}
        self.state_bounds: dict[str, tuple[float, float]] = {}
        self.interpolation_method: dict[str, tuple] = {}
        self.mode: dict[str, gp.MVar] = {}
        self.rules = ValidatorRules()
        self.gamma_length = self.milp.addMVar(self.N, name='system_gamma_length')

        # binary expansion option
        if n_gamma_digits is None:
            self.gamma_length_binary = None
        else:
            self.gamma_length_binary = [
                BinaryExpressionValue(
                    self.milp,
                    size=n_gamma_digits,
                    unit=gamma_unit_length,
                    name=f'gamma_length_binary[{i}]',
                )
                for i in range(N)
            ]
            self.milp.addConstrs(
                (
                    self.gamma_length[i].item() - self.gamma_length_binary[i].value
                    <= gamma_unit_length / 2
                    for i in range(N)
                ),
                name='gamma_diff_binary_ub',
            )
            self.milp.addConstrs(
                (
                    self.gamma_length[i] - self.gamma_length_binary[i].value >= 0
                    for i in range(N)
                ),
                name='gamma_diff_binary_lb',
            )

    def set_gamma_length(self, i: int, gamma_diff: gp.LinExpr | gp.Var) -> None:
        """
        Relate the length of `i`-th time step with some milp expression.
        `i` should be in [1, N].
        """
        self.milp.addConstr(self.gamma_length[i - 1] == gamma_diff)

    def add_state(
        self,
        name: str,
        lb: float,
        ub: float,
        qualitative: bool = False,
    ) -> gp.MVar:
        self.state[name] = self.milp.addMVar(
            self.N + 1,
            lb=lb,
            ub=ub,
            vtype=gp.GRB.CONTINUOUS if not qualitative else gp.GRB.INTEGER,
            name=name,
        )
        self.state_bounds[name] = (lb, ub)
        return self.state[name]

    def add_mode(
        self,
        name: str,
    ) -> gp.MVar:
        self.mode[name] = self.milp.addMVar(
            self.N + 1,
            vtype=gp.GRB.BINARY,
            name=name,
        )

        return self.mode[name]

    def add_dynamics(
        self,
        var: str,
        lb: float = 0.0,
        ub: float | None = None,
        mode: str | None = None,
        constant: bool = False,
    ):
        """
        Add dynamics of the variable.
        Currently only support linear dynamics.
        """
        if ub is None:
            ub = lb

        if constant:
            self.interpolation_method[var] = ('constant',)
        else:
            self.interpolation_method[var] = ('linear',)

        if mode is None:
            self.milp.addConstr(
                self.state[var][1:] - self.state[var][:-1] >= self.gamma_length * lb,
                name=f"dynamics_lb[mode=global][var={var}]",
            )
            self.milp.addConstr(
                self.state[var][1:] - self.state[var][:-1] <= self.gamma_length * ub,
                name=f"dynamics_ub[mode=global][var={var}]",
            )
            if not constant:
                self.rules.append(L.f(1.0, f'd_{var}') >= lb, tag='dynamics')
                self.rules.append(L.f(1.0, f'd_{var}') <= ub, tag='dynamics')
        else:
            self.milp.addConstrs(
                (
                    (self.mode[mode][i].item() == 1)
                    >> (
                        self.state[var][i + 1].item() - self.state[var][i].item()
                        >= self.gamma_length[i].item() * lb
                    )
                    for i in range(self.N)
                ),
                name=f"dynamics_lb[mode={mode}][var={var}]",
            )
            self.milp.addConstrs(
                (
                    (self.mode[mode][i].item() == 1)
                    >> (
                        self.state[var][i + 1].item() - self.state[var][i].item()
                        <= self.gamma_length[i].item() * ub
                    )
                    for i in range(self.N)
                ),
                name=f"dynamics_ub[mode={mode}][var={var}]",
            )

            if not constant:
                self.rules.append(
                    L.f(1.0, f'd_{var}') >= lb, lambda x: x[mode] == 1, tag='dynamics'
                )
                self.rules.append(
                    L.f(1.0, f'd_{var}') <= ub, lambda x: x[mode] == 1, tag='dynamics'
                )

    def add_double_integrator_dynamics(
        self,
        x: str,
        dx: str,
        ddx: str,
    ):
        assert self.gamma_length_binary is not None

        self.interpolation_method[dx] = ('linear',)
        self.interpolation_method[x] = ('integral', dx)

        self.milp.addConstr(self.state[ddx][-1].item() == self.state[ddx][-2].item())

        for i in range(self.N):
            self.milp.addConstr(
                self.state[dx][i + 1].item() - self.state[dx][i].item()
                == self.gamma_length_binary[i].multiply(
                    self.milp, self.state[ddx][i].item()
                )
            )
            self.milp.addConstr(
                self.state[x][i + 1].item() - self.state[x][i].item()
                == self.gamma_length_binary[i].multiply(
                    self.milp,
                    0.5 * (self.state[dx][i + 1].item() + self.state[dx][i].item()),
                )
            )

    def add_invariants(
        self,
        mode: str,
        *inequalities: LinearInequality,
    ):
        """
        Add invariants to the system.

        Args:
            mode: mode name
            inequalities: list of linear inequalities
        """
        for inequality_id, inequality in enumerate(inequalities):
            expr = inequality.to_gt_zero().expr.apply_to_variables(self.state)

            self.milp.addConstrs(
                (
                    (self.mode[mode][i].item() == 1) >> (expr[i].item() >= 0)
                    for i in range(self.N + 1)
                ),
                name=f"inv_backward[mode={mode}][{inequality_id}]",
            )
            self.milp.addConstrs(
                (
                    (self.mode[mode][i].item() == 1) >> (expr[i + 1].item() >= 0)
                    for i in range(self.N)
                ),
                name=f"inv_forward[mode={mode}][{inequality_id}]",
            )

            self.rules.append(inequality, lambda x: x[mode] == 1, tag='invariant')

    def add_jumps(
        self,
        two_modes: tuple[str, str],
        *guards: LinearInequality,
    ):
        """
        Add jump conditions between two modes.

        Args:
            two_modes: tuple of current mode and next mode
            guards: guard conditions, at the current mode
        """
        for inequality_id, inequality in enumerate(guards):
            expr = inequality.to_gt_zero().expr.apply_to_variables(self.state)[1:]

            # indicator[i] == 1 if and only if the current and next mode are active at time i
            indicator = self.milp.addVars(
                self.N,
                vtype=gp.GRB.BINARY,
                name=f"indicator[mode={two_modes}, inequality={inequality_id}]",
            )
            self.milp.addConstrs((
                indicator[i]
                == gp.and_(
                    self.mode[two_modes[0]][i].item(),
                    self.mode[two_modes[1]][i + 1].item(),
                )
                for i in range(self.N)
            ))

            self.milp.addConstrs(
                ((indicator[i] == 1) >> (expr[i].item() >= 0) for i in range(self.N)),
                name=f"jump[mode={two_modes[0]}->{two_modes[1]}][{inequality_id}]",
            )
            self.rules.append(
                inequality,
                lambda x, y: x[two_modes[1]] == 1 and y[two_modes[0]] == 1,
                tag='jump',
            )

    def set_initial_state(
        self,
        **kwargs: float | bool | tuple[float, float],
    ):
        for var, value in kwargs.items():
            if isinstance(value, float):
                self.milp.addConstr(
                    self.state[var][0].item() == value, name=f"initial_{var}"
                )
                self.rules.append(L.f(1.0, var) >= value, tag='initial')
                self.rules.append(L.f(1.0, var) <= value, tag='initial')
            elif isinstance(value, bool):
                self.milp.addConstr(
                    self.mode[var][0].item() == value, name=f"initial_{var}"
                )
                self.rules.append(L.f(1.0, var) >= value, tag='initial')
                self.rules.append(L.f(1.0, var) <= value, tag='initial')
            elif isinstance(value, tuple):
                self.milp.addRange(
                    self.state[var][0].item(), value[0], value[1], name=f"initial_{var}"
                )
                self.rules.append(L.f(1.0, var) >= value[0], tag='initial')
                self.rules.append(L.f(1.0, var) <= value[1], tag='initial')
            else:
                raise ValueError(f"Unknown type of initial state {var}: {value}")

    def validate(self, trace: Trace) -> bool:
        initial = self.rules.filter_by_tag('initial')
        initial.validate(trace.samples[0])

        dynamics = self.rules.filter_by_tag('dynamics')
        for sample in trace.samples_with_derivatives:
            dynamics.validate(sample)

        invariant = self.rules.filter_by_tag('invariant')
        for sample in trace.samples:
            invariant.validate(sample)

        jump = self.rules.filter_by_tag('jump')
        for sample, sample_before in zip(trace.samples[1:], trace.samples[:-1]):
            jump.validate(sample, sample_before)

        return True

    def get_trace(self, interpolation_interval: float | None = None) -> Trace:
        time = np.zeros(self.N + 1)
        time[1:] = np.cumsum(self.gamma_length.Xn)

        state = {}
        for var in self.state:
            state[var] = np.array(self.state[var].Xn)
        for var in self.mode:
            # regularize to exact 1.0 or 0.0
            raw = np.array(self.mode[var].Xn)
            state[var] = np.where(raw > 0.5, 1.0, 0.0)

        trace = Trace(time, state)

        if interpolation_interval is None:
            return trace

        # interpolation

        size = int((trace.time[-1] - trace.time[0]) / interpolation_interval) + 1

        constant_variables = [
            var
            for var, method in self.interpolation_method.items()
            if method[0] == 'constant'
        ]
        integral_variables = [
            var
            for var, method in self.interpolation_method.items()
            if method[0] == 'integral'
        ]
        linear_variables = [
            var
            for var in state
            if var not in constant_variables and var not in integral_variables
        ]  # default is linear

        trace_linear = trace.interpolate(
            np.linspace(trace.time[0], trace.time[-1], size),
            [*self.mode.keys(), *constant_variables],
            [*linear_variables, *integral_variables],
        )

        # compute interpolation for integral variables
        state_integral = {
            var: trace_linear[var]
            for var in trace.state
            if var not in integral_variables
        }
        time = trace_linear.time

        for var in state:
            if var not in integral_variables:
                state_integral[var] = trace_linear[var]

        for var in integral_variables:
            derivative_var = self.interpolation_method[var][1]
            state_integral[var] = (
                integrate.cumulative_trapezoid(
                    state_integral[derivative_var], time, initial=0
                )
                + trace[var][0]
            )

        return Trace(time, state_integral)



import gurobipy as gp
import json
from random import randint
from stltspref.problem import StlMilpProblem
from stltspref.benchmarks import get_benchmark
import matplotlib.pyplot as plt
from stltspref.linear_expression import *
from stltspref.stl import make_unique


def random_dist(prob: StlMilpProblem):
    """
    Creates a random objective function based on a Hamming distance 
    to a randomly chosen point in solution space.
    """
    obj = 0
    for var in prob.milp_variables.stl_sat.values():
        for v in var.values():
            if randint(0, 1):
                obj += 1-v
            else:
                obj += v
    return obj


def hamm_dist(prob: StlMilpProblem, coeffs: dict[gp.Var, int], constant: int):
    """
    Given a dictionary `coeffs` of previous coefficients of variables in problem `prob`,
    and a constant value (to ensure the objective is nonnegative),
    creates a new objective function based on a Hamming distance to the previous solutions.
    """
    obj = 0
    for var in prob.milp_variables.stl_sat.values():
        for v in var.values():
            if v.Xn == 1:
                constant += 1
                coeffs[v] -= 1 
            else:
                coeffs[v] += 1 
            obj += coeffs[v]*v
    return obj + constant, constant


def val_dist(prob: StlMilpProblem, i: int, milp: gp.Model):
    """
    Adds variables and constraints to encode absolute values
    and creates an objective function based on a value distance to the previous solutions.
    """
    obj = 0
    state = prob.system_model.state
    for var in state:
        for v in state[var].tolist():
            v_subs = milp.addVar(name=v.VarName+f"_subs_{i}", lb=-float('inf'))
            milp.addConstr(v_subs == v - v.Xn, name=v.VarName+f"_subs_{i}")
            v_abs = milp.addVar(name=v.VarName+f"_abs_{i}", lb=0)
            milp.addConstr(v_abs == gp.abs_(v_subs), name=v.VarName+f"_abs_{i}")
            obj += v_abs
    return obj


def encode_row(sample):
    """
    Puts trace values into rows for json encoding.
    """
    row = {}
    row['time'] = float(sample['t'])
    row['value'] = {k: float(v) for k, v in sample.items() if k != 't'}
    return row


def diversity_finder(
    prob: StlMilpProblem, 
    diversity_mode: str, 
    N: int, 
    numsols: int, 
    time_limit: float = 10000, 
    gap: float = 0.00001, 
    absolute_gap: float = 0.0001,
    export_traces: bool = False,
    export_path: str = ""
):
    """
    Generates `numsols` solutions to the STL trace 
    synthesis problem `prob` using a diversity heuristic.

    Args:
        `prob` (StlMilpProblem): The STL trace synthesis problem object,
            generated beforehand with system variables and a system model.
        `diversity_mode` (str): Identifier for the diversity method to use.
            Possible values are 
            - `'gurobi'` for the Solution Pool method,
            - `'hamming'` for the Hamming Distance Maximization method,
            - `'random'` for the Randomized Hamming Distance Maximization method,
            - `'value'` for the Value Distance Maximization method.
        `N` (int): Number of intervals of linearity in the generated traces.
        `numsols` (int): Number of different solutions to generate.
        `time_limit` (float): Maximum computation time in seconds.
        `gap` (float): Relative gap between incumbent solutions and
            known upper bounds to end trace search.
        `absolute_gap` (float): Absolute gap between incumbent solutions and
            known upper bounds to end trace search.
        `export_traces` (bool): Whether to export `json` files containing 
            the generated trace values.
        `export_path` (str): File path used when exporting trace value data.
    
    Raises:
        `ValueError`: Whenever `diversity_mode` is not one of its four possible values.
    """
    milp = prob.milp
    milp.Params.IntegralityFocus = 1
    # set chosen parameters
    milp.Params.TimeLimit = time_limit
    milp.Params.MIPGap = gap
    milp.Params.MIPGapAbs = absolute_gap
    if diversity_mode == "gurobi":
        prob.search_satisfaction(numsols, 2)
        if prob.has_solution:
            # get trace values, plot them and save them
            traces_continuous = prob.get_trace_result(interpolation=True)
            traces_raw = prob.get_trace_result(interpolation=False)
            for i in range(milp.SolCount):
                traces_continuous[i].df.plot(x='t')
                if export_traces:
                    rows = [encode_row(sample) for sample in traces_raw[i].samples]
                    with open(export_path+f'output_{i+1}.json', 'w') as f:
                        json.dump(rows, f, indent=2)
            plt.show()
        else:
            print(f'No trace found with N {N}')

    elif diversity_mode == "hamming" or diversity_mode == "value":
        prob.search_satisfaction(1, 0)
        if prob.has_solution:
            # get trace values, plot them and save them
            traces_continuous = prob.get_trace_result(interpolation=True)
            traces_continuous[0].df.plot(x='t')
            if export_traces:
                rows = [encode_row(sample) for sample in prob.get_trace_result(interpolation=False)[0].samples]
                with open(export_path+f'output_{1}.json', 'w') as f:
                    json.dump(rows, f, indent=2)
        obj = 0
        if diversity_mode == "value":
            # ensure to have the same time sequence
            g = prob.get_gamma_result()
            milp.addConstrs(
                (prob.milp_variables.gamma[i] == g[i] for i in range(prob.N+1)), 
                name="gamma_equality"
            )
        else:
            # initialize coeffs and constant value in hamming objective 
            coeffs = {v: 0 for var in prob.milp_variables.stl_sat.values() for v in var.values()}
            constant = 0
        for sol in range(1, numsols):
            # update objective
            if diversity_mode == "hamming":
                obj, constant = hamm_dist(prob, coeffs, constant)
            else:
                obj += val_dist(prob, sol, milp)
            milp.setObjective(obj, gp.GRB.MAXIMIZE)
            # search sol
            prob.search_satisfaction(1, 0)
            if prob.has_solution:
                # get trace values, plot them and save them
                traces_continuous = prob.get_trace_result(interpolation=True)
                traces_continuous[0].df.plot(x='t')
                if export_traces:
                    rows = [encode_row(sample) for sample in prob.get_trace_result(interpolation=False)[0].samples]
                    with open(export_path+f'output_{sol+1}.json', 'w') as f:
                        json.dump(rows, f, indent=2)
        plt.show()

    elif diversity_mode == "random":
        for sol in range(numsols):
            # create a random objective
            obj = random_dist(prob)
            milp.setObjective(obj, gp.GRB.MAXIMIZE)
            # solve
            prob.search_satisfaction(1, 0)
            if prob.has_solution:
                # get trace values, plot them and save them
                traces_continuous = prob.get_trace_result(interpolation=True)
                traces_continuous[0].df.plot(x='t')
                if export_traces:
                    rows = [encode_row(sample) for sample in prob.get_trace_result(interpolation=False)[0].samples]
                    with open(export_path+f'output_{sol+1}.json', 'w') as f:
                        json.dump(rows, f, indent=2)
        plt.show()
    
    else:
        raise ValueError(f"Unsupported method name: {diversity_mode}.")


def pn_pair_finder(
    benchmark_name: str,
    N: int, 
    time_limit: float = 10000, 
    gap: float = 0.00001, 
    absolute_gap: float = 0.0001,
    export_traces: bool = False,
    export_path: str = ""
):
    """
    Generates two solutions to two STL trace 
    synthesis problems, one with a positive spec and one negative.

    Args:
        `benchmark_name` (str): Identifier for the benchmark to use.
            Possible values are
            - `'dstop'`,
            - `'rnc1'` to `'rnc3'`,
            - `'nav1'` or `'nav2'`,
            - `'iso1'` to `'iso8'`.
        `N` (int): Number of intervals of linearity in the generated traces.
        `numsols` (int): Number of different solutions to generate.
        `time_limit` (float): Maximum computation time in seconds.
        `gap` (float): Relative gap between incumbent solutions and
            known upper bounds to end trace search.
        `absolute_gap` (float): Absolute gap between incumbent solutions and
            known upper bounds to end trace search.
        `export_traces` (bool): Whether to export `json` files containing 
            the generated trace values.
        `export_path` (str): File path used when exporting trace value data.
    """
    milp = gp.Model()
    milp.Params.IntegralityFocus = 1
    # set chosen parameters
    milp.Params.TimeLimit = time_limit
    milp.Params.MIPGap = gap
    milp.Params.MIPGapAbs = absolute_gap
    # initialize problem
    prob = get_benchmark(milp, benchmark_name, N=N, delta=0.1)
    milp.update()
    # rename all positive vars and constrs in milp to distinguish pos and neg vars
    for var in milp.getVars():
        var.VarName += "_posver"
    for constr in milp.getConstrs():
        constr.ConstrName += "_posver"
    # initialize the negative problem
    neg_spec = make_unique(prob.stl_spec.negation())
    prob2 = get_benchmark(milp, benchmark_name, N=N, custom_spec=neg_spec, delta=0.1)
    milp.update()
    # rename all negative vars and constrs in milp for clarity
    for var in milp.getVars():
        if not var.VarName.endswith("_posver"):
            var.VarName += "_negver"
    for constr in milp.getConstrs():
        if not constr.ConstrName.endswith("_posver"):
            constr.ConstrName += "_negver"
    # create the objective function (adding absolute value constraints)
    obj = 0
    milp.update()
    for v in prob.state:
        for i in range(prob.N + 1):
            var1 = prob.state[v][i].item()
            var2 = prob2.state[v][i].item()
            v_sub = milp.addVar(name=v+"_sub")
            milp.addConstr(v_sub == var1 - var2, name=v+f"[{i}]_sub")
            v_abs = milp.addVar(name=v+f"[{i}]_abs")
            milp.addConstr(v_abs == gp.abs_(v_sub), name=v+f"[{i}]_abs")
            obj += v_abs
    # add constraints to have identical time sequences in pos and neg problems
    milp.addConstrs(
        (prob.milp_variables.gamma[i] == prob2.milp_variables.gamma[i] for i in range(prob.N+1)),
        name="gamma_equality"
    )
    milp.setObjective(obj, gp.GRB.MINIMIZE)
    # solve prob2 (or prob1, it uses the same milp solve)
    prob2.search_satisfaction()
    if prob2.has_solution:
        # get trace values, plot them and save them for positive prob
        prob.get_trace_result(interpolation=True)[0].df.plot(x='t')
        if export_traces:
            rows = [encode_row(sample) for sample in prob.get_trace_result(interpolation=False)[0].samples]
            with open(export_path+f'output_pos.json', 'w') as f:
                json.dump(rows, f, indent=2)
        # same for negative prob, in case they differ
        prob2.get_trace_result(interpolation=True)[0].df.plot(x='t')
        if export_traces:
            rows = [encode_row(sample) for sample in prob.get_trace_result(interpolation=False)[0].samples]
            with open(export_path+f'output_neg.json', 'w') as f:
                json.dump(rows, f, indent=2)
        plt.show()


def benchmark_pref_synth(
    benchmark_name: str, 
    diversity_mode: str, 
    N: int, 
    numsols: int,
    **kwargs
):
    """
    Uses preferential synthesis method identified by `diversity_mode`
    on the benchmark identified by `benchmark_name`.

    Args:
        `benchmark_name` (str): Identifier for the benchmark to use.
            Possible values are
            - `'dstop'`,
            - `'rnc1'` to `'rnc3'`,
            - `'nav1'` or `'nav2'`,
            - `'iso1'` to `'iso8'`.
        `diversity_mode` (str): Identifier for the diversity method to use.
            Possible values are 
            - `'gurobi'` for the Solution Pool method,
            - `'hamming'` for the Hamming Distance Maximization method,
            - `'random'` for the Randomized Hamming Distance Maximization method,
            - `'value'` for the Value Distance Maximization method,
            - `'pn_pair'` for the Positive-Negative Pair method.
        `N` (int): Number of intervals of linearity in the generated traces.
        `numsols` (int): Number of different solutions to generate.
        `time_limit` (float): Maximum computation time in seconds. 
            Defaults to `10000`.
        `gap` (float): Relative gap between incumbent solutions and
            known upper bounds to end trace search. Defaults to `0.00001`.
        `absolute_gap` (float): Absolute gap between incumbent solutions and
            known upper bounds to end trace search. Defaults to `0.0001`.
        `export_traces` (bool): Whether to export `json` files containing 
            the generated trace values. Defaults to `False`.
        `export_path` (str): File path used when exporting trace value data. 
            Defaults to `''`.
    """
    if diversity_mode == "pn_pair":
        pn_pair_finder(benchmark_name, N, **kwargs)
    else:
        milp = gp.Model()
        prob = get_benchmark(milp, benchmark_name, N=N, delta=0.1)
        diversity_finder(prob, diversity_mode, N, numsols, **kwargs)



from __future__ import annotations

import numpy as np

from .trace import Trace


class StlFormula:
    """STL formula."""

    def __init__(self, name: str, children: list[StlFormula] | None = None):
        self.name = name
        self.bounded = False
        self.interval = None
        self._timing = None

        if children is None:
            self.children = []
        else:
            self.children = children

    def __hash__(self) -> int:
        return hash(self.name)

    @property
    def shortname(self) -> str:
        return self.name[:200]

    @property
    def child(self) -> StlFormula:
        assert len(self.children) == 1
        return self.children[0]

    def get_subformula(self, name: str) -> StlFormula | None:
        for phi in self.get_subformulas():
            if phi.name == name:
                return phi
        return None

    def get_subformulas(self, depth: int | None = None) -> list[StlFormula]:
        if depth is not None and depth == 0:
            return [self]
        if len(self.children) == 0:
            return [self]
        else:
            subformulas = []
            for child in self.children:
                candidates = child.get_subformulas(
                    depth=depth - 1 if depth is not None else None
                )
                for p in candidates:
                    if p.name not in [q.name for q in subformulas]:
                        subformulas.append(p)
            return [*subformulas, self]

    def __repr__(self) -> str:
        return self.name

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        """Return the robustness value of the formula at each step of the trace."""
        raise NotImplementedError

    def negation(self) -> StlFormula:
        raise NotImplementedError



from __future__ import annotations

import numpy as np
import pandas as pd
import scipy


class Trace:
    state: dict[str, np.ndarray]
    time: np.ndarray

    def __init__(self, time, state: dict[str, np.ndarray]):
        self.state = state
        self.time = time

    @classmethod
    def from_df(cls, df: pd.DataFrame) -> Trace:
        return cls(
            df['t'].to_numpy(),
            {var: df[var].to_numpy() for var in df.columns if var != 't'},
        )

    def __getitem__(self, item):
        return self.state[item]

    @property
    def df(self) -> pd.DataFrame:
        return pd.DataFrame(self.samples)

    def at(self, time: float, var: str) -> float:
        return np.interp(time, self.time, self.state[var]).item()

    def _derivative(self, var: str, i: int) -> float:
        assert 0 <= i < len(self.time) - 1
        return (self.state[var][i + 1] - self.state[var][i]) / (
            self.time[i + 1] - self.time[i]
        ).item()

    @property
    def samples(self) -> list[dict[str, float | bool]]:
        return [
            {
                't': self.time[i],
                **{var: self.state[var][i].item() for var in self.state},
            }
            for i in range(len(self.time))
        ]

    @property
    def samples_with_derivatives(self) -> list[dict[str, float | bool]]:
        return [
            {
                't': self.time[i],
                **{var: self.state[var][i].item() for var in self.state},
                **{
                    f'd_{var}': self._derivative(var, i)
                    for var in self.state
                    if self.state[var].dtype == np.float64
                },
            }
            for i in range(len(self.time) - 1)
            if self.time[i] != self.time[i + 1]  # avoid zero division
        ]

    def interpolate(
        self,
        steps: np.ndarray | None = None,
        constant_states: list[str] | None = None,
        linear_states: list[str] | None = None,
    ) -> Trace:
        if steps is None:
            steps = np.arange(self.time[0], self.time[-1], 0.1)
        if constant_states is None:
            constant_states = []
        if linear_states is None:
            linear_states = [
                state for state in self.state.keys() if state not in constant_states
            ]

        # merge and remove duplicates (tiny difference is considered as duplicate)
        time = np.concatenate((steps, self.time))
        time.sort(kind='mergesort')
        time = time[np.diff(time, append=np.inf) >= 1e-9]

        state = {}
        for var in linear_states:
            state[var] = np.interp(time, self.time, self.state[var])
        for var in constant_states:
            state[var] = scipy.interpolate.interp1d(
                self.time, self.state[var], kind='previous'
            )(time)
        return Trace(time, state)



from __future__ import annotations

from typing import Callable

from .linear_expression import LinearInequality

EPSILON = 1e-3


class Approx:
    def __init__(self, value: float, epsilon: float = EPSILON):
        self.epsilon = epsilon
        self.value = value

    def __eq__(self, other):
        return abs(other - self.value) <= self.epsilon

    def __le__(self, other):
        return self.value - other <= self.epsilon

    def __ge__(self, other):
        return self.value - other >= -self.epsilon


class ApproxRange:
    def __init__(self, lb, ub, epsilon: float = EPSILON):
        self.epsilon = epsilon
        self.lb = lb
        self.ub = ub

    def __eq__(self, other):
        return self.lb - self.epsilon <= other <= self.ub + self.epsilon


class ValidatorRules:
    rules: list[tuple[LinearInequality, Callable | None, str | None]]

    def __init__(self, rules: list | None = None):
        if rules is None:
            rules = []
        self.rules = rules

    def append(
        self, inequality: LinearInequality, condition=None, tag: str | None = None
    ):
        self.rules.append((inequality, condition, tag))

    def filter_by_tag(self, tag: str) -> ValidatorRules:
        return ValidatorRules([x for x in self.rules if x[2] == tag])

    def validate(self, input_1, input_optional=None) -> None:
        for inequality, condition, _ in self.rules:
            if condition is None:
                triggered = True
            elif input_optional is None:
                triggered = condition(input_1)
            else:
                triggered = condition(input_1, input_optional)

            if triggered:
                value = inequality.to_gt_zero().expr.eval(input_1)
                assert value >= Approx(
                    0
                ), f"{inequality.name} violated  with value {value}, at {input_1}"



from __future__ import annotations

import dataclasses
import functools
from typing import TYPE_CHECKING

import gurobipy as gp
import numpy as np

import stltspref.stl as stl
from stltspref.milp_helper import add_and_aux_var, add_or_aux_var, at_least_one
from stltspref.stl_atomic import Predicate

from stltspref.constants import EPSILON

if TYPE_CHECKING:
    Vars = gp.tupledict[int, gp.Var]


@dataclasses.dataclass
class MilpVariables:
    N: int
    gamma: Vars
    stl_sat: dict[stl.StlFormula, Vars]
    pred_sat: dict[Predicate, Vars]
    pred_strong_sat: dict[Predicate, Vars]
    sat_durations: dict[stl.BoundedAlw, Vars]
    unsat_durations: dict[stl.BoundedEv, Vars]

    def get_values(self) -> dict:
        def get_values_from_vars(vars: Vars) -> np.ndarray:
            return np.array([vars[i].X for i in vars])

        def get_bool_values_from_vars(vars: Vars) -> np.ndarray:
            return np.array([True if vars[i].X >= 0.5 else False for i in vars])

        return dict(
            gamma=get_values_from_vars(self.gamma),
            stl_sat={
                phi: get_bool_values_from_vars(vars)
                for phi, vars in self.stl_sat.items()
            },
            pred_sat={
                p: get_bool_values_from_vars(self.pred_sat[p]) for p in self.pred_sat
            },
            sat_durations={
                phi: get_values_from_vars(dur)
                for phi, dur in self.sat_durations.items()
            },
            unsat_durations={
                phi: get_values_from_vars(dur)
                for phi, dur in self.unsat_durations.items()
            },
        )


def make_milp_variables(
    milp: gp.Model,
    spec: stl.StlFormula,
    N: int,
    gamma_0: float,
    gamma_N: float,
    gamma_unit_length: float = 0.01,
) -> MilpVariables:
    subformulas = spec.get_subformulas()
    predicates = [phi.predicate for phi in subformulas if isinstance(phi, stl.Atomic)]

    gamma = milp.addVars(N + 1, lb=gamma_0, ub=gamma_N, name='gamma')

    stl_sat = {
        phi: milp.addVars(
            list(range(1, N + 1)), vtype=gp.GRB.BINARY, name=f'stl_{phi.shortname}'
        )
        for phi in subformulas
    }
    pred_sat = {
        p: milp.addVars(
            list(range(0, N + 1)), vtype=gp.GRB.BINARY, name=f'pred_sat_{p.name}'
        )
        for p in predicates
    }
    pred_strong_sat = {
        p: milp.addVars(
            list(range(0, N + 1)), vtype=gp.GRB.BINARY, name=f'pred_strong_sat_{p.name}'
        )
        for p in predicates
    }
    sat_durations = {
        phi: milp.addVars(
            list(range(1, N + 1)),
            lb=0,
            ub=gamma_N - gamma_0,
            name=f'sat_dur_{phi.shortname}',
        )
        for phi in subformulas
        if isinstance(phi, stl.BoundedAlw)
    }
    unsat_durations = {
        phi: milp.addVars(
            list(range(1, N + 1)),
            lb=0,
            ub=gamma_N - gamma_0,
            name=f'unsat_dur_{phi.shortname}',
        )
        for phi in subformulas
        if isinstance(phi, stl.BoundedEv)
    }

    for phi in subformulas:
        if isinstance(phi, stl.BoundedAlw) or isinstance(phi, stl.BoundedEv):
            assert phi.interval[0] < phi.interval[1]
            assert 0 <= phi.interval[0]
            assert phi.interval[1] <= gamma_N - gamma_0 - EPSILON

    return MilpVariables(
        N=N,
        gamma=gamma,
        stl_sat=stl_sat,
        pred_sat=pred_sat,
        pred_strong_sat=pred_strong_sat,
        sat_durations=sat_durations,
        unsat_durations=unsat_durations,
    )


def add_timeline_constraints(
    milp: gp.Model,
    v: MilpVariables,
    gamma_0: float,
    gamma_N: float,
    gamma_unit_length: float = 0.01,
) -> None:
    N = v.N
    gamma = v.gamma

    gamma_length = {i: gamma[i] - gamma[i - 1] for i in range(1, N + 1)}

    milp.addConstr(gamma[0] == gamma_0, name='gamma_0')
    milp.addConstr(gamma[N] == gamma_N, name='gamma_N')

    milp.addConstrs(
        (gamma_length[i] >= EPSILON for i in range(1, N + 1)),
        name='gamma_monotonicity',
    )


def add_predicate_constraints(
    milp: gp.Model,
    v: MilpVariables,
    system_states: dict[str, gp.MVar],
    delta: float = 0.1,
) -> None:
    N = v.N
    pred_sat = v.pred_sat
    pred_strong_sat = v.pred_strong_sat
    predicates = list(pred_sat.keys())

    for p in predicates:
        postfix = f'{p.name}'
        expr_value = p.expr_value(system_states)
        depth = expr_value / p.inequality.expr.norm

        milp.addConstrs(
            (
                pred_sat[p][i - 1] >= pred_strong_sat[p][i] - pred_strong_sat[p][i - 1]
                for i in range(1, N + 1)
            ),
            name=f'stationary_rise_{postfix}',
        )
        milp.addConstrs(
            (
                pred_sat[p][i + 1] >= pred_strong_sat[p][i] - pred_strong_sat[p][i + 1]
                for i in range(N)
            ),
            name=f'stationary_fall_{postfix}',
        )
        milp.addConstrs(
            ((pred_sat[p][i] == 1) >> (depth[i].item() >= 0) for i in range(N + 1)),
            name=f'pred_sat_geq_{postfix}',
        )
        milp.addConstrs(
            (
                (pred_sat[p][i] == 0) >> (depth[i].item() <= -EPSILON)
                for i in range(N + 1)
            ),
            name=f'pred_sat_lt_{postfix}',
        )
        milp.addConstrs(
            (
                (pred_strong_sat[p][i] == 1) >> (depth[i].item() >= delta)
                for i in range(N + 1)
            ),
            name=f'pred_strong_sat_geq_{postfix}',
        )
        milp.addConstrs(
            (
                (pred_strong_sat[p][i] == 0) >> (depth[i].item() <= delta - EPSILON)
                for i in range(N + 1)
            ),
            name=f'pred_strong_sat_lt_{postfix}',
        )
        milp.addConstrs(
            (pred_sat[p][i] >= pred_strong_sat[p][i] for i in range(N + 1)),
            name=f'strong_sat_then_sat_{postfix}',
        )  # redundant


### STL constraints


def add_full_stl_constraints(
    milp: gp.Model,
    v: MilpVariables,
    root: stl.StlFormula,
    require_satisfaction=True,
) -> None:
    root_subformulas = root.get_subformulas()
    known_formulas = list(v.stl_sat.keys())

    for phi in root_subformulas:
        if phi not in known_formulas:
            raise ValueError(f"Unknown subformula {phi} in the STL formula")

    subformulas = [
        phi for phi in known_formulas if phi in root_subformulas
    ]  # only consider known formulas that are in the root formula

    for phi in subformulas:
        add_stl_constraints(phi, milp, v)
    if require_satisfaction:
        milp.addConstr(v.stl_sat[root][1] == 1, name='root_satisfaction')


@functools.singledispatch
def add_stl_constraints(
    phi: stl.StlFormula,
    milp: gp.Model,
    v: MilpVariables,
) -> None:
    raise NotImplementedError(f"Unsupported STL formula {phi}")


@add_stl_constraints.register
def _(phi: stl.Top, milp: gp.Model, v: MilpVariables):
    phi_sat = v.stl_sat[phi]
    milp.addConstrs((phi_sat[i] == 1 for i in phi_sat), name=f'stl[top]')


@add_stl_constraints.register
def _(phi: stl.Bottom, milp: gp.Model, v: MilpVariables):
    phi_sat = v.stl_sat[phi]
    milp.addConstrs((phi_sat[i] == 0 for i in phi_sat), name=f'stl[bottom]')


@add_stl_constraints.register
def _(phi: stl.And, milp: gp.Model, v: MilpVariables):
    phi_sat = v.stl_sat[phi]
    children_sat = [v.stl_sat[child] for child in phi.children]
    def children_sat_and(i):
        return gp.and_(*[child[i] for child in children_sat])
    milp.addConstrs(
        (phi_sat[i] == children_sat_and(i) for i in phi_sat),
        name=f'stl[and]_{phi.shortname}',
    )


@add_stl_constraints.register
def _(phi: stl.Or, milp: gp.Model, v: MilpVariables):
    phi_sat = v.stl_sat[phi]
    children_sat = [v.stl_sat[child] for child in phi.children]
    def children_sat_or(i):
        return gp.or_(*[child[i] for child in children_sat])
    milp.addConstrs(
        (phi_sat[i] == children_sat_or(i) for i in phi_sat),
        name=f'stl[or]_{phi.shortname}',
    )


@add_stl_constraints.register
def _(phi: stl.Until, milp: gp.Model, v: MilpVariables):
    postfix = f'{phi.shortname}'
    N = v.N
    z1 = milp.addVars(list(range(1, N)), vtype=gp.GRB.BINARY, name=f'z1_{postfix}')
    z2 = milp.addVars(list(range(1, N)), vtype=gp.GRB.BINARY, name=f'z2_{postfix}')
    milp.addConstr(
        v.stl_sat[phi][N] == gp.and_(v.stl_sat[phi.phi1][N], v.stl_sat[phi.phi2][N]),
        name=f'stl[until]_N_{postfix}',
    )
    for i in range(1, N):
        z1[i] = add_and_aux_var(
            milp,
            v.stl_sat[phi][i + 1],
            v.stl_sat[phi.phi1][i],
            name=f'aux1_{postfix}[{i}]',
        )
        z2[i] = add_and_aux_var(
            milp,
            v.stl_sat[phi.phi1][i],
            v.stl_sat[phi.phi2][i],
            name=f'aux2_{postfix}[{i}]',
        )
        milp.addConstr(
            v.stl_sat[phi][i] == gp.or_(z1[i], z2[i]), name=f'stl_until_{postfix}[{i}]'
        )


@add_stl_constraints.register
def _(phi: stl.Release, milp: gp.Model, v: MilpVariables):
    postfix = f'{phi.shortname}'
    N = v.N
    z1 = milp.addVars(list(range(1, N)), vtype=gp.GRB.BINARY, name=f'z1_{postfix}')
    z2 = milp.addVars(list(range(1, N)), vtype=gp.GRB.BINARY, name=f'z2_{postfix}')
    milp.addConstr(
        v.stl_sat[phi][N] == gp.or_(v.stl_sat[phi.phi1][N], v.stl_sat[phi.phi2][N]),
        name=f'stl[release]_N_{postfix}',
    )
    for i in range(1, N):
        z1[i] = add_or_aux_var(
            milp,
            v.stl_sat[phi][i + 1],
            v.stl_sat[phi.phi1][i],
            name=f'aux1_{postfix}[{i}]',
        )
        z2[i] = add_or_aux_var(
            milp,
            v.stl_sat[phi.phi1][i],
            v.stl_sat[phi.phi2][i],
            name=f'aux2_{postfix}[{i}]',
        )
        milp.addConstr(
            v.stl_sat[phi][i] == gp.and_(z1[i], z2[i]),
            name=f'stl[release]_{postfix}[{i}]',
        )


@add_stl_constraints.register
def _(phi: stl.Alw, milp: gp.Model, v: MilpVariables):
    phi_sat = v.stl_sat[phi]
    postfix = f'{phi.shortname}'
    N = v.N
    milp.addConstr(phi_sat[N] == v.stl_sat[phi.child][N], name=f'stl[alw]_N_{postfix}')
    milp.addConstrs(
        (
            phi_sat[i] == gp.and_(phi_sat[i + 1], v.stl_sat[phi.child][i])
            for i in range(1, N)
        ),
        name=f'stl[alw]_{postfix}',
    )


@add_stl_constraints.register
def _(phi: stl.Ev, milp: gp.Model, v: MilpVariables):
    phi_sat = v.stl_sat[phi]
    postfix = f'{phi.shortname}'
    N = v.N
    milp.addConstr(phi_sat[N] == v.stl_sat[phi.child][N], name=f'stl[ev]_N_{postfix}')
    milp.addConstrs(
        (
            phi_sat[i] == gp.or_(phi_sat[i + 1], v.stl_sat[phi.child][i])
            for i in range(1, N)
        ),
        name=f'stl[ev]_{postfix}',
    )


@add_stl_constraints.register
def _(phi: stl.Atomic, milp: gp.Model, v: MilpVariables):
    phi_sat = v.stl_sat[phi]
    pred_strong_sat = v.pred_strong_sat[phi.predicate]
    milp.addConstrs(
        (
            phi_sat[i] == gp.or_(pred_strong_sat[i - 1], pred_strong_sat[i])
            for i in phi_sat
        ),
        name=f'stl[atomic]_{phi.shortname}',
    )


@add_stl_constraints.register
def _(phi: stl.BoundedAlw, milp: gp.Model, v: MilpVariables):
    _bounded_alw_constraints(
        milp=milp,
        phi_sat=v.stl_sat[phi],
        child_sat=v.stl_sat[phi.child],
        milp_prefix=f'stl[alwI]_{phi.shortname}',
        sat_durations=v.sat_durations[phi],
        interval=phi.interval,
        gamma=v.gamma,
    )


@add_stl_constraints.register
def _(phi: stl.BoundedEv, milp: gp.Model, v: MilpVariables):
    phi_sat = v.stl_sat[phi]
    child_sat = v.stl_sat[phi.child]

    phi_unsat = milp.addVars(
        phi_sat.keys(), vtype=gp.GRB.BINARY, name=f'unsat_{phi.shortname}'
    )
    milp.addConstrs(
        (phi_unsat[i] == 1 - phi_sat[i] for i in phi_sat), name=f'unsat_{phi.shortname}'
    )
    child_unsat = milp.addVars(
        child_sat.keys(), vtype=gp.GRB.BINARY, name=f'unsat_child_{phi.shortname}'
    )
    milp.addConstrs(
        (child_unsat[i] == 1 - child_sat[i] for i in child_sat),
        name=f'unsat_child_{phi.shortname}',
    )

    _bounded_alw_constraints(
        milp=milp,
        phi_sat=phi_unsat,
        child_sat=child_unsat,
        milp_prefix=f'stl[evI]_{phi.shortname}',
        sat_durations=v.unsat_durations[phi],
        interval=phi.interval,
        gamma=v.gamma,
    )


def _bounded_alw_constraints(
    milp: gp.Model,
    phi_sat: Vars,
    child_sat: Vars,
    milp_prefix: str,
    sat_durations: Vars,
    interval: tuple[float, float],
    gamma: Vars,
) -> None:
    """
    Suppose Alw_I(phi) is strongly satisfactory in Gamma[i].
    For any interval J that intersects with Gamma[i] + I, phi must strongly satisfactory in J.
    """
    ## Prepare variables
    N = len(phi_sat)
    a, b = interval
    gamma_length = {i: gamma[i] - gamma[i - 1] for i in range(1, N + 1)}

    _indicator: dict[tuple[str, int, str, int], gp.Var] = {}

    def _indicate(kind, i, op, j) -> gp.Var:
        if (kind, i, op, j) in _indicator:
            return _indicator[kind, i, op, j]
        assert kind in ['inf', 'sup']
        assert op in ['gt', 'lt', 'geq', 'leq']
        assert 0 <= i <= N or (i == N + 1 and kind == 'inf')
        assert 0 <= j <= N

        name = f'{milp_prefix}_indicate_[{kind}_{i}_{op}_{j}]'

        z = _indicator[kind, i, op, j] = milp.addVar(vtype=gp.GRB.BINARY, name=name)

        if kind == 'inf':
            lhs = gamma[i - 1] + interval[0]
        elif kind == 'sup':
            lhs = gamma[i] + interval[1]

        if op == 'gt':
            milp.addConstr(
                (z == 1) >> (lhs >= gamma[j] + EPSILON),
            )
        elif op == 'lt':
            milp.addConstr(
                (z == 1) >> (lhs <= gamma[j] - EPSILON),
            )
        elif op == 'geq':
            milp.addConstr(
                (z == 1) >> (lhs >= gamma[j]),
            )
        elif op == 'leq':
            milp.addConstr(
                (z == 1) >> (lhs <= gamma[j]),
            )
        return z

    ## Constraints for |R_i|
    # R_1 = 0 if <phi>_1 = 0 else Gamma[1]
    # R_i = 0 if <phi>_i = 0 else Gamma[i] + R_{i-1} for i = [2, N]
    milp.addConstrs(
        ((child_sat[i] == 0) >> (sat_durations[i] == 0) for i in range(1, N + 1)),
        name=f'{milp_prefix}_sat_dur_unsat',
    )

    _length = lambda i: (0 if i == 1 else sat_durations[i - 1]) + gamma_length[i]  # type: ignore
    milp.addConstrs(
        (
            (child_sat[i] == 1) >> (sat_durations[i] == _length(i))
            for i in range(1, N + 1)
        ),
        name=f'{milp_prefix}_sat_dur_sat',
    )

    ## Constraints for <Alw_I psi>_i
    for i in range(1, N + 1):
        milp_prefix_i = f"{milp_prefix}[{i}]"
        # Rule `Alw_I_pos_1`:
        #   <Alw_I psi>_i = 1, intersect(Gamma[i] + I, Gamma[j]) ==> <psi>_j = 1
        #   for j = [i, N]
        milp.addConstrs(
            (
                at_least_one(
                    1 - phi_sat[i],
                    _indicate('sup', i, 'leq', j - 1),
                    _indicate('inf', i, 'gt', j),
                    child_sat[j],
                )
                for j in range(i, N + 1)
            ),
            name=f"{milp_prefix_i}_pos_1",
        )
        # Rule `Alw_I_pos_1N`:
        #   <Alw_I psi>_i = 1, intersect(Gamma[i] + I, (gamma[N], +inf)) ==> <psi>_N = 1
        milp.addConstr(
            at_least_one(1 - phi_sat[i], _indicate('sup', i, 'lt', N), child_sat[N]),
            name=f"{milp_prefix_i}_pos_1N",
        )
        # Rule `Alw_I_neg_1`:
        #   <Alw_I psi>_i = 0, gamma[i - 1] + b <= gamma[j] <= gamma[i] + b ==> |R_j| < |I|
        #   for j = [i, N]
        z = milp.addVars(
            ['neg_1', 'neg_2', 'neg_3'],
            range(i, N + 1),
            vtype=gp.GRB.BINARY,
            name=f'{milp_prefix_i}_z',
        )
        milp.addConstrs(
            (
                at_least_one(
                    phi_sat[i],
                    _indicate('sup', i - 1, 'geq', j),
                    _indicate('sup', i, 'leq', j),
                    z['neg_1', j],
                )
                for j in range(i, N + 1)
            ),
            name=f"{milp_prefix_i}_neg_1",
        )
        milp.addConstrs((
            (z['neg_1', j] == 1) >> (sat_durations[j] <= (b - a))
            for j in range(i, N + 1)
        ))

        # Rule `Alw_I_neg_2`:
        #   <Alw_I psi>_i = 0, gamma[j - 1] <= gamma[i] + b <= gamma[j]  ==> |R_j| < gamma[j] - gamma[i] - a
        #   for j = [i + 1, N]
        milp.addConstrs(
            (
                at_least_one(
                    phi_sat[i],
                    _indicate('sup', i, 'lt', j - 1),
                    _indicate('sup', i, 'gt', j),
                    z['neg_2', j],
                )
                for j in range(i + 1, N + 1)
            ),
            name=f"{milp_prefix_i}_neg_2",
        )

        milp.addConstrs((
            (z['neg_2', j] == 1) >> (sat_durations[j] <= gamma[j] - gamma[i] - a)
            for j in range(i + 1, N + 1)
        ))

        # Rule `Alw_I_neg_3`:
        #   <Alw_I psi>_i = 0, gamma[i] + a <= gamma[N] <= gamma[i] + b ==> |R_N| < gamma[N] - gamma[i] - a
        milp.addConstr(
            at_least_one(
                phi_sat[i],
                _indicate('inf', i + 1, 'gt', N),
                _indicate('sup', i, 'lt', N),
                z['neg_3', N],
            ),
            name=f"{milp_prefix_i}_neg_3",
        )
        milp.addConstr(
            (z['neg_3', N] == 1) >> (sat_durations[N] <= (gamma[N] - gamma[i] - a))
        )

        # Rule `Alw_I_neg_4`:
        #   <Alw_I psi>_i = 0, gamma[N] <= gamma[i] + a  ==> <psi>_N = 0
        milp.addConstr(
            at_least_one(
                phi_sat[i], _indicate('inf', i + 1, 'leq', N), 1 - child_sat[N]
            ),
            name=f"{milp_prefix_i}_neg_4",
        )



from __future__ import annotations

import gurobipy as gp
import numpy as np

from .linear_expression import LinearInequality
from .stl_base import StlFormula
from .trace import Trace


class Predicate:
    """Linear inequality predicate."""

    def __init__(
        self,
        inequality: LinearInequality,
        qualitative: bool = False,
    ) -> None:
        self.name = inequality.name
        self.inequality = inequality
        self.qualitative = qualitative

    def __hash__(self) -> int:
        return hash(self.name)

    def __repr__(self) -> str:
        return self.name

    def eval_depth(self, state: dict[str, np.ndarray]) -> np.ndarray:
        """Evaluate the (signed) depth at the given state."""
        value = self.inequality.expr.eval(state) * self.inequality.sign
        return value / self.inequality.expr.norm

    def expr_value(self, state: dict[str, gp.MVar]) -> gp.MLinExpr:
        """Return the value of the expression at the given state."""
        return self.inequality.expr.apply_to_variables(state) * self.inequality.sign


class Atomic(StlFormula):
    def __init__(self, symbol: LinearInequality, qualitative: bool = False):
        super().__init__(symbol.name)
        self.predicate = Predicate(symbol, qualitative=qualitative)

    def negation(self) -> StlFormula:
        if not self.predicate.qualitative:
            return Atomic(self.predicate.inequality.invert())
        else:
            # negation of x >= 0 is x <= -1
            expr = self.predicate.inequality.to_gt_zero().expr
            return Atomic(expr <= -1, qualitative=True)

    def robust_semantics(self, trace: Trace, memo=None) -> np.ndarray:
        """Return the robustness value of the formula at each step of the trace."""
        if memo is not None and self.name in memo:
            return memo[self.name]
        return self.predicate.eval_depth(trace.state)



from __future__ import annotations

import gurobipy as gp


def reactionTime():
    return 0.6


def amax():
    return 5


def bmin():
    return 6


def bmax():
    return 10


def amax_y():
    return 1.5


def bmin_y():
    return 1.5


def add_square(milp: gp.Model, v: gp.MVar):
    assert v.ndim == 1
    sq = milp.addMVar(shape=v.shape)
    for i in range(v.size):
        milp.addGenConstrPoly(
            v[i].item(),
            sq[i].item(),
            (1, 0, 0),
            options='FuncPieces=-1 FuncPieceError=0.01',
        )
    return sq


def preRssDistance(
    milp: gp.Model,
    rear_vel: gp.MVar,
    rear_vel_sq: gp.MVar,
    front_vel_sq: gp.MVar,
    name: str,
):
    preRssDistance = milp.addMVar(
        shape=rear_vel.shape, lb=-gp.GRB.INFINITY, name=f'preRssDistance_{name}'
    )
    milp.addConstr(
        preRssDistance
        == (
            reactionTime() * rear_vel
            + 0.5 * amax() * reactionTime() ** 2
            + (
                rear_vel_sq
                + 2 * amax() * reactionTime() * rear_vel
                + (amax() * reactionTime()) ** 2
            )
            / (2 * bmin())
            - (front_vel_sq) / (2 * bmax())
        )
    )
    return preRssDistance


def preLateralRssDistance(
    milp: gp.Model,
    left_vel: gp.MVar,
    left_vel_sq: gp.MVar,
    right_vel: gp.MVar,
    right_vel_sq: gp.MVar,
    name: str,
):
    # left is bigger
    preLateralRssDistance = milp.addMVar(
        shape=left_vel.shape, lb=-gp.GRB.INFINITY, name=f'preLateralRssDistance_{name}'
    )

    milp.addConstr(
        preLateralRssDistance
        == (
            (left_vel - right_vel) * reactionTime()
            + amax_y() * reactionTime() ** 2
            + (
                left_vel_sq
                + 2 * reactionTime() * amax_y() * left_vel
                + (reactionTime() * amax_y()) ** 2
                + right_vel_sq
                - 2 * reactionTime() * amax_y() * right_vel
                + (reactionTime() * amax_y()) ** 2
            )
            / (2 * bmin_y())
        )
    )
    return preLateralRssDistance


def setRssDistance(
    milp: gp.Model,
    a_var: gp.MVar,
    b_var: gp.MVar,
    a_vel: gp.MVar,
    b_vel: gp.MVar,
):
    # Precompute squares
    a_vel_sq = add_square(milp, a_vel)
    b_vel_sq = add_square(milp, b_vel)

    # 1) car a is rear
    preRssDistance_a = preRssDistance(
        milp,
        rear_vel=a_vel,
        rear_vel_sq=a_vel_sq,
        front_vel_sq=b_vel_sq,
        name='car_a',
    )
    milp.addConstrs(
        a_var[i].item() == gp.max_(preRssDistance_a[i].item(), constant=0.0)
        for i in range(a_var.shape[0])
    )

    # 2) car b is rear
    preRssDistance_b = preRssDistance(
        milp,
        rear_vel=b_vel,
        rear_vel_sq=b_vel_sq,
        front_vel_sq=a_vel_sq,
        name='car_b',
    )
    milp.addConstrs(
        b_var[i].item() == gp.max_(preRssDistance_b[i].item(), constant=0.0)
        for i in range(b_var.shape[0])
    )


def setLateralRssDistance(
    milp: gp.Model,
    a_var: gp.MVar,
    b_var: gp.MVar,
    a_yvel: gp.MVar,
    b_yvel: gp.MVar,
):
    # Precompute squares
    a_vel_sq = add_square(milp, a_yvel)
    b_vel_sq = add_square(milp, b_yvel)

    # 1) car a is right
    preLateralRssDistance_a = preLateralRssDistance(
        milp,
        right_vel=a_yvel,
        left_vel=b_yvel,
        right_vel_sq=a_vel_sq,
        left_vel_sq=b_vel_sq,
        name='car_a',
    )
    milp.addConstrs(
        a_var[i].item() == gp.max_(preLateralRssDistance_a[i].item(), constant=0.0)
        for i in range(a_var.shape[0])
    )

    # 1) car b is right
    preLateralRssDistance_b = preLateralRssDistance(
        milp,
        right_vel=b_yvel,
        left_vel=a_yvel,
        right_vel_sq=b_vel_sq,
        left_vel_sq=a_vel_sq,
        name='car_b',
    )
    milp.addConstrs(
        b_var[i].item() == gp.max_(preLateralRssDistance_b[i].item(), constant=0.0)
        for i in range(b_var.shape[0])
    )



from __future__ import annotations

import typing

from .linear_expression import LinearExpression as L
from .stl import Alw, And, Atomic, Bottom, BoundedAlw, BoundedEv, Ev, Or, Top, Until

Real = typing.Any

car_width = 1.7
car_length = 4.7

extension = True


# longitudinal danger: long. car distance <= car length + RSS distance,


def danger_x(
    ax: Real, ax_vel: Real, bx: Real, bx_vel: Real, length_a: Real, length_b: Real
):
    formulas = [
        ((bx + -ax) + -(L.unit('rssDistance_a') + length_b) <= 0) @ 'danger_car_a_rear',
        ((ax + -bx) + -(L.unit('rssDistance_b') + length_a) <= 0) @ 'danger_car_b_rear',
    ]
    return And(*[Atomic(f) for f in formulas])


# //lateral danger: lat. car distance <= car width + RSS distance
# //1.6m = width of smart. Numbers taken from GA-RSS paper,
# //(except for ay_max (see paper in notes),
# //by_min, by_max (derived by trigonometry))
def danger_y(
    ay: Real, ay_vel: Real, by: Real, by_vel: Real, width_a: Real, width_b: Real
):
    formulas = [
        ((by + -ay) + -(L.unit('lateralRssDistance_a') + width_b) <= 0)
        @ 'danger_car_a_right',
        ((ay + -by) + -(L.unit('lateralRssDistance_b') + width_a) <= 0)
        @ 'danger_car_b_right',
    ]
    return And(*[Atomic(f) for f in formulas])


# //danger : RSS distance violated in both lat. and long. direction
# //e.g. if cars are driving next to eachother in different lanes,
# //there is no danger even when ax=bx if lateral safety distance is kept
def danger(
    ax: Real,
    ax_vel: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    by: Real,
    by_vel: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
):
    return And(
        danger_x(ax, ax_vel, bx, bx_vel, length_a, length_b),
        danger_y(ay, ay_vel, by, by_vel, width_a, width_b),
    )


# //safety = not danger
def safe(
    ax: Real,
    ax_vel: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    by: Real,
    by_vel: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
):
    return (
        danger(
            ax,
            ax_vel,
            ay,
            ay_vel,
            bx,
            bx_vel,
            by,
            by_vel,
            length_a,
            width_a,
            length_b,
            width_b,
        )
    ).negation()


# //----defining road structure----
#
# // i = 0, 1, 2 is a lane number
# // lane centers are at 1.75, 5.25, 8.75
def atLane(laneID: Real, i: int):
    if laneID.variables[0] == 'lane_a':
        car = 'a'
    elif laneID.variables[0] == 'lane_b':
        car = 'b'
    else:
        raise ValueError(f"Unknown laneID {laneID}")

    position = L.unit(f'{car}y')

    # 0--3.5, 3.5--7, 7--10.5
    return And(
        Atomic(position <= 3.5 * (i + 1) + car_width),
        Atomic(position >= 3.5 * i),
    )


# //OLD: the point masses ax and bx are at least one car length apart
# //NEW: b is slightly ahead of a (for use in cut-in/out scenario)
def aheadOf(ax: Real, bx: Real):
    return Atomic(ax + -bx <= 0)


def strictAheadOf(ax: Real, bx: Real, b_length: Real):
    return Atomic(ax + b_length + -bx <= 0)


def safe_dur(
    ax: Real,
    ax_vel: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    by: Real,
    by_vel: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
    durness: float = 0.6,
):
    safety_formula = safe(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        length_a,
        width_a,
        length_b,
        width_b,
    )
    if durness == 0.0:
        return safety_formula
    else:
        return BoundedAlw(
            (0.0, durness),
            safety_formula,
        )


def danger_dur(
    ax: Real,
    ax_vel: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    by: Real,
    by_vel: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
    durness: float = 0.6,
):
    danger_formula = danger(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        length_a,
        width_a,
        length_b,
        width_b,
    )
    if durness == 0.0:
        return danger_formula
    else:
        return BoundedAlw(
            (0.0, durness),
            danger_formula,
        )


def adjacentLanes(lane1: Real, lane2: Real):
    if isinstance(lane1, float) and isinstance(lane2, float):
        if abs(lane1 - lane2) == 1:
            return Top()
        else:
            return Bottom()
    else:
        return NotImplementedError()
    return Or(
        And(
            Atomic(lane1 + -lane2 >= -1, qualitative=True),
            Atomic(lane1 + -lane2 <= -1, qualitative=True),
        ),
        And(
            Atomic(lane1 + -lane2 >= 1, qualitative=True),
            Atomic(lane1 + -lane2 <= 1, qualitative=True),
        ),
    )


def sameLane(lane1: Real, lane2: Real):
    if isinstance(lane1, float) and isinstance(lane2, float):
        if lane1 - lane2 == 0:
            return Top()
        else:
            return Bottom()
    else:
        return NotImplementedError()
    return And(
        Atomic(lane1 + -lane2 >= 0, qualitative=True),
        Atomic(lane1 + -lane2 <= 0, qualitative=True),
    )


def enteringLane(lane_a: Real, targetLane: Real):
    return And((atLane(lane_a, targetLane)).negation(), Ev(atLane(lane_a, targetLane)))


def leavingLane(lane_a: Real, startLane: Real):
    return And(atLane(lane_a, startLane), Ev(atLane(lane_a, startLane).negation()))


def accel(
    x: Real,
    x_vel: Real,
    x_acc: Real,
    y: Real,
    refX: Real,
    refX_vel: Real,
    laneID: Real,
    startLane: Real,
    rate: float = 0.1,
):
    if extension:
        f1 = Or(Atomic(x_acc >= rate), Atomic(x_vel + -refX_vel >= rate))
    else:
        f1 = Atomic(x_vel + -refX_vel >= rate)
    f2 = atLane(laneID, startLane)
    return And(f1, f2)


def decel(
    x: Real,
    x_vel: Real,
    x_acc: Real,
    y: Real,
    refX: Real,
    refX_vel: Real,
    laneID: Real,
    startLane: Real,
    rate: float = 0.1,
):
    if extension:
        f1 = Or(Atomic(x_acc <= -rate), Atomic(x_vel + -refX_vel <= -rate))
    else:
        f1 = Atomic(x_vel + -refX_vel <= -rate)
    f2 = atLane(laneID, startLane)
    return And(f1, f2)


# // SV and POV b:  b is behind a and they are in the same lane until there is danger (because b is accelerating)
def scenario3(
    ax: Real,
    ax_vel: Real,
    ax_acc: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    bx_acc: Real,
    by: Real,
    by_vel: Real,
    lane_a: Real,
    initialLane_a: Real,
    lane_b: Real,
    initialLane_b: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
):
    if extension:
        f1 = Or(aheadOf(bx, ax), strictAheadOf(bx, ax, length_a))
    else:
        f1 = strictAheadOf(bx, ax, length_a)
    # // probably needed without restrictions on ax
    f2 = safe_dur(
        bx,
        bx_vel,
        by,
        by_vel,
        ax,
        ax_vel,
        ay,
        ay_vel,
        length_b,
        width_b,
        length_a,
        width_a,
    )
    f3 = Or(
        adjacentLanes(initialLane_a, initialLane_b),
        sameLane(initialLane_a, initialLane_b),
    )
    f4 = Until(
        And(
            atLane(lane_a, initialLane_a),
            accel(bx, bx_vel, bx_acc, by, ax, ax_vel, lane_b, initialLane_b),
        ),
        (
            danger_dur(
                ax,
                ax_vel,
                ay,
                ay_vel,
                bx,
                bx_vel,
                by,
                by_vel,
                length_a,
                width_a,
                length_b,
                width_b,
            )
        ),
    )  # // implicitly because of danger: their distance will decrease
    return And(f1, f2, f3, f4)


# // SV and POV b:  a is behind b and they are in the same lane until there is danger (because b is braking)
def scenario4(
    ax: Real,
    ax_vel: Real,
    ax_acc: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    bx_acc: Real,
    by: Real,
    by_vel: Real,
    lane_a: Real,
    initialLane_a: Real,
    lane_b: Real,
    initialLane_b: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
):
    if extension:
        f1 = Or(aheadOf(ax, bx), strictAheadOf(ax, bx, length_b))
    else:
        f1 = strictAheadOf(ax, bx, length_b)
    f2 = safe_dur(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        length_a,
        width_a,
        length_b,
        width_b,
    )
    f3 = Or(
        adjacentLanes(initialLane_a, initialLane_b),
        sameLane(initialLane_a, initialLane_b),
    )
    f4 = Until(
        And(
            atLane(lane_a, initialLane_a),
            decel(bx, bx_vel, bx_acc, by, ax, ax_vel, lane_b, initialLane_b),
        ),
        (
            danger_dur(
                ax,
                ax_vel,
                ay,
                ay_vel,
                bx,
                bx_vel,
                by,
                by_vel,
                length_a,
                width_a,
                length_b,
                width_b,
            )
        ),
    )  # // implicitly because of danger: their distance will decrease
    return And(f1, f2, f3, f4)


# // SV a and POV b: a is in  lane j, b in lane i, and a wants to merge in front of b, but they get too close to each other
# // because of the arrow in the table: set b to acceleration?
def scenario7(
    ax: Real,
    ax_vel: Real,
    ax_acc: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    bx_acc: Real,
    by: Real,
    by_vel: Real,
    lane_a: Real,
    initialLane_a: Real,
    lane_b: Real,
    initialLane_b: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
):
    if extension:
        f1 = Or(aheadOf(bx, ax), strictAheadOf(bx, ax, length_a))
    else:
        f1 = strictAheadOf(bx, ax, length_a)
    # // probably needed without restrictions on ax

    f2 = safe_dur(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        length_a,
        width_a,
        length_b,
        width_b,
    )
    f3 = enteringLane(lane_a, initialLane_b)
    f4 = Until(
        (accel(bx, bx_vel, bx_acc, by, ax, ax_vel, lane_b, initialLane_b)),
        danger_dur(
            ax,
            ax_vel,
            ay,
            ay_vel,
            bx,
            bx_vel,
            by,
            by_vel,
            length_a,
            width_a,
            length_b,
            width_b,
        ),
    )  # // implicitly because of danger: their distance will decrease
    return And(f1, f2, f3, f4)


# // SV a and POV b: a is behind b, b is decelerating while a is changing lanes
def scenario8(
    ax: Real,
    ax_vel: Real,
    ax_acc: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    bx_acc: Real,
    by: Real,
    by_vel: Real,
    lane_a: Real,
    initialLane_a: Real,
    lane_b: Real,
    initialLane_b: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
    alpha=1.0,
):
    if extension:
        f1 = Or(aheadOf(ax, bx), strictAheadOf(ax, bx, length_b))
    else:
        f1 = strictAheadOf(ax, bx, length_b)

    f2 = safe_dur(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        length_a,
        width_a,
        length_b,
        width_b,
    )
    if initialLane_a != initialLane_b:
        raise ValueError("initialLane_a != initialLane_b")
    # f3 = sameLane(initialLane_a, initialLane_b)
    f4 = atLane(lane_a, initialLane_a)
    f5 = leavingLane(lane_a, initialLane_a)
    f6 = Until(
        decel(
            bx,
            bx_vel,
            bx_acc,
            by,
            ax,
            ax_vel,
            lane_b,
            initialLane_b,
            rate=alpha,
        ),
        (
            danger_dur(
                ax,
                ax_vel,
                ay,
                ay_vel,
                bx,
                bx_vel,
                by,
                by_vel,
                length_a,
                width_a,
                length_b,
                width_b,
            )
        ),
    )  #    // implicitly because of danger: their distance will decrease
    return And(f1, f2, f4, f5, f6)


# //Predicates for cut-in/out


# //cutIn - TODO: make durness-friendly
def cutIn(
    ax: Real,
    ax_vel: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    by: Real,
    by_vel: Real,
    lane_a: Real,
    initialLane_a: Real,
    lane_b: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
):
    if extension:
        f = []
    else:
        f = [strictAheadOf(ax, bx, length_b)]
    ev_content = And(
        danger_dur(
            ax,
            ax_vel,
            ay,
            ay_vel,
            bx,
            bx_vel,
            by,
            by_vel,
            length_a,
            width_a,
            length_b,
            width_b,
        ),
        BoundedEv(
            (0.0, 0.6),
            And(
                *f,
                atLane(lane_b, initialLane_a),
            ),
        ),
    )
    return And(
        atLane(lane_a, initialLane_a),
        (atLane(lane_b, initialLane_a)).negation(),
        Ev(ev_content),
    )


# //three vehicle cut-out
# //def cutOut_sameLane(ax: Real,ay: Real,bx: Real,by: Real,cx: Real,cy: Real, lane_a: Real, lane_b: Real, lane_c: Real, length_a: Real, width_a: Real, length_b: Real, width_b: Real, length_c: Real, width_c: Real, initialLane_a: Real) : Bool =
# //  atLane(lane_a, lane_c) /\  atLane(lane_b,lane_c) /\ atLane(lane_c,initialLane_a)
# //  /\ aheadOf(ax,bx) /\ aheadOf(bx,cx)
# //  /\ ((atLane(lane_a,initialLane_a) /\ atLane(lane_c,initialLane_a)) U (~(atLane(lane_b,initialLane_a)) /\ danger(ax,ay,cx,cy, length_a, width_a, length_c, width_c))) // a and c both keep lane until b has left lane and danger with c occurs
# //  /\ (leavingLane(lane_b,initialLane_a)) // b leaves lane eventually


# //two vehicle cut-out does not have an extension
def cutOut(
    ax: Real,
    ax_vel: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    by: Real,
    by_vel: Real,
    lane_a: Real,
    initialLane_a: Real,
    lane_b: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
):
    return And(
        atLane(lane_a, initialLane_a),
        leavingLane(lane_b, initialLane_a),
        Ev(
            And(
                danger_dur(
                    ax,
                    ax_vel,
                    ay,
                    ay_vel,
                    bx,
                    bx_vel,
                    by,
                    by_vel,
                    length_a,
                    width_a,
                    length_b,
                    width_b,
                ),
                BoundedEv((0.0, 0.6), atLane(lane_b, initialLane_a).negation()),
            )
        ),
    )


#  // this is possibly already an extension: no extension would be to include an aheadOf


def scenario1(
    ax: Real,
    ax_vel: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    by: Real,
    by_vel: Real,
    lane_a: Real,
    initialLane_a: Real,
    lane_b: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
):
    f1 = safe_dur(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        length_a,
        width_a,
        length_b,
        width_b,
    )
    f2 = Until(
        atLane(lane_a, initialLane_a),
        danger_dur(
            ax,
            ax_vel,
            ay,
            ay_vel,
            bx,
            bx_vel,
            by,
            by_vel,
            length_a,
            width_a,
            length_b,
            width_b,
        ),
    )  # SV lane keep until danger

    f3 = cutIn(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        lane_a,
        initialLane_a,
        lane_b,
        length_a,
        width_a,
        length_b,
        width_b,
    )  # POV cut-in
    return And(f1, f2, f3)


# //--scenario 5: SV a changes lane i -> j, POV b cuts in by changing lane -> i
def scenario5(
    ax: Real,
    ax_vel: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    by: Real,
    by_vel: Real,
    lane_a: Real,
    initialLane_a: Real,
    lane_b: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
):
    f1 = safe_dur(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        length_a,
        width_a,
        length_b,
        width_b,
    )
    f2 = leavingLane(lane_a, initialLane_a)
    f3 = cutIn(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        lane_a,
        initialLane_a,
        lane_b,
        length_a,
        width_a,
        length_b,
        width_b,
    )
    return And(f1, f2, f3)


# //--scenario 6: SV a changes lane i -> j, POV b changes lane i -> j dangerously
def scenario6(
    ax: Real,
    ax_vel: Real,
    ay: Real,
    ay_vel: Real,
    bx: Real,
    bx_vel: Real,
    by: Real,
    by_vel: Real,
    lane_a: Real,
    initialLane_a: Real,
    lane_b: Real,
    length_a: Real,
    width_a: Real,
    length_b: Real,
    width_b: Real,
):
    f1 = safe_dur(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        length_a,
        width_a,
        length_b,
        width_b,
    )
    f2 = leavingLane(lane_a, initialLane_a)
    f3 = cutOut(
        ax,
        ax_vel,
        ay,
        ay_vel,
        bx,
        bx_vel,
        by,
        by_vel,
        lane_a,
        initialLane_a,
        lane_b,
        length_a,
        width_a,
        length_b,
        width_b,
    )
    return And(f1, f2, f3)


def get_scenario(id: int, alpha=0.1):
    if id == 1:
        return scenario1(
            L.unit('ax'),
            L.unit('ax_vel'),
            #    L.unit('ax_acc'),
            L.unit('ay'),
            L.unit('ay_vel'),
            #    L.unit('ay_acc'),
            L.unit('bx'),
            L.unit('bx_vel'),
            #    L.unit('bx_acc'),
            L.unit('by'),
            L.unit('by_vel'),
            #    L.unit('by_acc'),
            lane_a=L.unit('lane_a'),
            initialLane_a=2.0,
            lane_b=L.unit('lane_b'),
            #   initialLane_b=initial_lane_b,
            length_a=car_length,
            width_a=car_width,
            length_b=car_length,
            width_b=car_width,
        )
    elif id == 3:
        return scenario3(
            L.unit('ax'),
            L.unit('ax_vel'),
            L.unit('ax_acc'),
            L.unit('ay'),
            L.unit('ay_vel'),
            #    L.unit('ay_acc'),
            L.unit('bx'),
            L.unit('bx_vel'),
            L.unit('bx_acc'),
            L.unit('by'),
            L.unit('by_vel'),
            #    L.unit('by_acc'),
            lane_a=L.unit('lane_a'),
            initialLane_a=1.0,
            lane_b=L.unit('lane_b'),
            initialLane_b=1.0,
            length_a=car_length,
            width_a=car_width,
            length_b=car_length,
            width_b=car_width,
        )
    elif id == 4:
        return scenario4(
            L.unit('ax'),
            L.unit('ax_vel'),
            L.unit('ax_acc'),
            L.unit('ay'),
            L.unit('ay_vel'),
            #    L.unit('ay_acc'),
            L.unit('bx'),
            L.unit('bx_vel'),
            L.unit('bx_acc'),
            L.unit('by'),
            L.unit('by_vel'),
            #    L.unit('by_acc'),
            lane_a=L.unit('lane_a'),
            initialLane_a=1.0,
            lane_b=L.unit('lane_b'),
            initialLane_b=1.0,
            length_a=car_length,
            width_a=car_width,
            length_b=car_length,
            width_b=car_width,
        )
    elif id == 5:
        return scenario5(
            L.unit('ax'),
            L.unit('ax_vel'),
            #    L.unit('ax_acc'),
            L.unit('ay'),
            L.unit('ay_vel'),
            #    L.unit('ay_acc'),
            L.unit('bx'),
            L.unit('bx_vel'),
            #   L.unit('bx_acc'),
            L.unit('by'),
            L.unit('by_vel'),
            #    L.unit('by_acc'),
            lane_a=L.unit('lane_a'),
            initialLane_a=1.0,
            lane_b=L.unit('lane_b'),
            length_a=car_length,
            width_a=car_width,
            length_b=car_length,
            width_b=car_width,
        )
    elif id == 6:
        return scenario6(
            L.unit('ax'),
            L.unit('ax_vel'),
            #    L.unit('ax_acc'),
            L.unit('ay'),
            L.unit('ay_vel'),
            #    L.unit('ay_acc'),
            L.unit('bx'),
            L.unit('bx_vel'),
            #   L.unit('bx_acc'),
            L.unit('by'),
            L.unit('by_vel'),
            #    L.unit('by_acc'),
            lane_a=L.unit('lane_a'),
            initialLane_a=1.0,
            lane_b=L.unit('lane_b'),
            length_a=car_length,
            width_a=car_width,
            length_b=car_length,
            width_b=car_width,
        )
    elif id == 7:
        return scenario7(
            L.unit('ax'),
            L.unit('ax_vel'),
            L.unit('ax_acc'),
            L.unit('ay'),
            L.unit('ay_vel'),
            #    L.unit('ay_acc'),
            L.unit('bx'),
            L.unit('bx_vel'),
            L.unit('bx_acc'),
            L.unit('by'),
            L.unit('by_vel'),
            #    L.unit('by_acc'),
            lane_a=L.unit('lane_a'),
            initialLane_a=1.0,
            lane_b=L.unit('lane_b'),
            initialLane_b=2.0,
            length_a=car_length,
            width_a=car_width,
            length_b=car_length,
            width_b=car_width,
        )
    elif id == 8:
        return scenario8(
            L.unit('ax'),
            L.unit('ax_vel'),
            L.unit('ax_acc'),
            L.unit('ay'),
            L.unit('ay_vel'),
            #    L.unit('ay_acc'),
            L.unit('bx'),
            L.unit('bx_vel'),
            L.unit('bx_acc'),
            L.unit('by'),
            L.unit('by_vel'),
            #    L.unit('by_acc'),
            lane_a=L.unit('lane_a'),
            initialLane_a=1.0,
            lane_b=L.unit('lane_b'),
            initialLane_b=1.0,
            length_a=car_length,
            width_a=car_width,
            length_b=car_length,
            width_b=car_width,
            alpha=alpha,
        )
    else:
        raise ValueError(f"Unknown scenario id {id}")


def get_danger(margin=0.0):
    return danger(
        L.unit('ax'),
        L.unit('ax_vel'),
        L.unit('ay'),
        L.unit('ay_vel'),
        L.unit('bx'),
        L.unit('bx_vel'),
        L.unit('by'),
        L.unit('by_vel'),
        length_a=car_length - margin,
        width_a=car_width - margin,
        length_b=car_length - margin,
        width_b=car_width - margin,
    )



from __future__ import annotations

import gurobipy as gp

from stltspref.problem import StlMilpProblem, create_stl_milp_problem
from stltspref.linear_expression import LinearExpression as L
from stltspref.milp_helper import ensure_monotonicity_by_derivative
from stltspref.stl import (
    Alw,
    And,
    Atomic,
    BoundedAlw,
    BoundedEv,
    BoundedUntil,
    Ev,
    Implies,
    Or,
    Until,
    StlFormula,
    make_unique,
)


def get_benchmark(milp: gp.Model, name: str, **kwargs) -> StlMilpProblem:
    if name.startswith('rnc'):
        return get_chasing_car(milp, spec=name, **kwargs)
    elif name == "dstop":
        return get_dstop(milp, **kwargs)
    elif name.startswith('nav'):
        return get_robot_navigation(milp, spec=name, **kwargs)
    elif name.startswith('iso'):
        return get_iso_rss(milp, spec=name, **kwargs)
    else:
        raise ValueError(f'Unknown benchmark: {name}')


def get_chasing_car(
    milp: gp.Model,
    N: int = 10,
    spec: str = 'rnc1',
    custom_spec: StlFormula | None = None,
    delta=0.1,
) -> StlMilpProblem:
    prob = create_stl_milp_problem(
        milp,
        N=N,
        delta=delta,
        gamma_N=20.0,
        gamma_unit_length=0.005,
        use_binary_expansion=True,
    )

    # System Dynamics
    car = prob.create_system_model()

    x1 = car.add_state('x1', 0, 500)
    v1 = car.add_state('v1', 0, 27)
    a1 = car.add_state('a1', -3, 3)

    x2 = car.add_state('x2', 0, 500)
    v2 = car.add_state('v2', 0, 27)
    a2 = car.add_state('a2', -3, 3)

    car.set_initial_state(
        x1=(45.0, 100.0),
        x2=(0.0, 5.0),
    )

    # # Dynamics
    car.add_dynamics('a1', -3, 3, constant=True)
    car.add_dynamics('a2', -3, 3, constant=True)

    car.add_double_integrator_dynamics('x1', 'v1', 'a1')
    car.add_double_integrator_dynamics('x2', 'v2', 'a2')

    # # Predicates in linear inequality form
    distance = L.f(1.0, 'x1', -1.0, 'x2')

    danger = (distance <= 10) @ 'danger'
    precedence = (distance >= 0.0) @ 'precedence'
    accelerating = (L.unit('a2') >= 1) @ 'accelerating'

    ensure_monotonicity_by_derivative(
        milp, [v2[i].item() - v1[i].item() for i in range(prob.N + 1)], prob.N
    )

    # Stl Spec
    if custom_spec is None:
        if spec == 'rnc1' or spec == 'rnc3':
            if spec == 'rnc1':
                safe_t = 1
            elif spec == 'rnc3':
                safe_t = 0.2
            sticky_accel = BoundedAlw((0, safe_t), Atomic(accelerating))
            control_spec = (
                Atomic(precedence),
                Implies(
                    Ev(Atomic(danger)),
                    Until(sticky_accel, Atomic(danger)),
                ),
            )

            scenario_spec = (BoundedEv((0, 9), BoundedAlw((0, 1), Atomic(danger))),)
            stl_spec = make_unique(
                And(
                    Alw(And(*control_spec)),
                    *scenario_spec,
                )
            )
        elif spec == 'rnc2':
            control_spec = (Atomic(precedence),)

            scenario_spec = (
                BoundedEv(
                    (0, 9),
                    And(
                        BoundedAlw((0, 1), Atomic(accelerating)),
                        BoundedAlw((0, 1), Atomic(danger)),
                        BoundedEv((1, 5), Atomic(danger).negation()),
                    ),
                ),
            )

            stl_spec = make_unique(
                And(
                    Alw(And(*control_spec)),
                    *scenario_spec,
                )
            )
        else:
            raise ValueError(f'Unknown spec: {spec}')
    else:
        stl_spec = custom_spec
    prob.initialize_milp_formulation(stl_spec)
    return prob


def get_dstop(
    milp: gp.Model,
    N: int = 10,
    custom_spec: StlFormula | None = None,
    delta=0.1,
) -> StlMilpProblem:
    prob = create_stl_milp_problem(
        milp,
        N=N,
        delta=delta,
        gamma_N=20.0,
        gamma_unit_length=0.005,
        use_binary_expansion=True,
    )

    # System Dynamics
    car = prob.create_system_model()

    x1 = car.add_state('x1', 0, 500)
    v1 = car.add_state('v1', 0, 27)
    a1 = car.add_state('a1', -3, 3)

    x2 = car.add_state('x2', 0, 500)
    v2 = car.add_state('v2', 0, 27)
    a2 = car.add_state('a2', -3, 3)

    car.set_initial_state(
        x1=(45.0, 100.0),
        x2=(0.0, 5.0),
    )

    # Dynamics
    car.add_dynamics('a1', -3, 3, constant=True)
    car.add_dynamics('a2', -3, 3, constant=True)

    car.add_double_integrator_dynamics('x1', 'v1', 'a1')
    car.add_double_integrator_dynamics('x2', 'v2', 'a2')

    # Predicates in linear inequality form
    distance = L.f(1.0, 'x1', -1.0, 'x2')

    danger = (distance <= 10) @ 'danger'
    precedence = (distance >= 0.0) @ 'precedence'
    safe = (distance >= 40) @ 'safe'

    ensure_monotonicity_by_derivative(
        milp, [v2[i].item() - v1[i].item() for i in range(prob.N + 1)], prob.N
    )

    # Stl Spec
    if custom_spec is None:
        stl_spec = make_unique(
            And(
                Alw(Atomic(precedence)),
                BoundedUntil(
                    (10, 15), 
                    BoundedEv((0, 10), Atomic(danger)),
                    Or(
                        Alw(Atomic(L.unit('v1') <= 0.1)), 
                        Alw(Atomic(L.unit('v2') <= 0.1))
                    )
                ),
                BoundedAlw((0,2), Atomic(safe))
            )
        )
    else:
        stl_spec = custom_spec
    prob.initialize_milp_formulation(stl_spec)
    return prob


def get_robot_navigation(
    milp: gp.Model,
    N: int = 10,
    spec: str = 'nav1',
    custom_spec: StlFormula | None = None,
    delta=0.01,
) -> StlMilpProblem:
    horizon = 40.0

    prob = create_stl_milp_problem(
        milp,
        N=N,
        delta=delta,
        gamma_N=horizon,
        gamma_unit_length=0.01,
        use_binary_expansion=False,
    )

    car = prob.create_system_model()

    x = car.add_state('x', 0, 10)
    y = car.add_state('y', 0, 10)

    modes = [
        LOC1 := 'loc1',
        LOC2 := 'loc2',
        LOC3 := 'loc3',
        LOC4 := 'loc4',
    ]

    for mode in modes:
        car.add_mode(mode)

    car.set_initial_state(
        x=(0, 3),
        y=0.0,
    )

    # Mode constraints should be manually maintained
    car.milp.addConstr(
        sum(car.mode[var] for var in modes) == 1,
        name='mode_exact_one',
    )

    # Dynamics
    car.add_dynamics('x', 1, mode=LOC1)
    car.add_dynamics('y', 0.1, 2, mode=LOC1)
    car.add_invariants(
        LOC1,
        L.unit('x') >= 0,
        L.unit('x') <= 5,
        L.unit('y') >= 5,
        L.unit('y') <= 10,
    )

    car.add_dynamics('x', 0.1, 2, mode=LOC2)
    car.add_dynamics('y', -1, mode=LOC2)
    car.add_invariants(
        LOC2,
        L.unit('x') >= 5,
        L.unit('x') <= 10,
        L.unit('y') >= 4,
        L.unit('y') <= 10,
    )

    car.add_dynamics('x', -1, mode=LOC3)
    car.add_dynamics('y', -2, -0.1, mode=LOC3)
    car.add_invariants(
        LOC3,
        L.unit('x') >= 5,
        L.unit('x') <= 10,
        L.unit('y') >= 0,
        L.unit('y') <= 4,
    )

    car.add_dynamics('x', -2, -0.1, mode=LOC4)
    car.add_dynamics('y', 1, mode=LOC4)
    car.add_invariants(
        LOC4,
        L.unit('x') >= 0,
        L.unit('x') <= 5,
        L.unit('y') >= 0,
        L.unit('y') <= 5,
    )

    goal_region = And(
        Atomic(L.unit('x') >= 4.0),
        Atomic(L.unit('x') <= 6.0),
        Atomic(L.unit('y') >= 2.0),
        Atomic(L.unit('y') <= 5.0),
    )

    unsafe_region = And(Atomic(L.unit('x') >= 9.0))

    in_l1 = And(
        Atomic(L.unit('x') <= 5.0),
        Atomic(L.unit('y') >= 5.0),
    )
    in_l3 = And(
        Atomic(L.unit('x') >= 5.0),
        Atomic(L.unit('y') <= 4.0),
    )

    in_l4 = And(
        Atomic(L.unit('x') <= 5.0),
        Atomic(L.unit('y') <= 5.0),
    )

    if custom_spec is None:
        if spec == 'nav1':
            scenario_spec = [
                Ev(BoundedAlw((0, 3), goal_region)),
                Ev(unsafe_region).negation(),
            ]
        elif spec == 'nav2':
            p = Atomic(L.unit('x') <= 5.0)
            scenario_spec = [
                Ev(in_l3),
                Alw(Implies(in_l3, BoundedEv((0, 3), p))),
            ]
        else:
            raise ValueError(f'Unknown spec: {spec}')

        stl_spec = make_unique(
            And(
                #  Alw(And(*control_spec)),
                *scenario_spec,
            )
        )
    else:
        stl_spec = custom_spec
    prob.initialize_milp_formulation(stl_spec)

    return prob


def get_iso_rss(
    milp: gp.Model,
    N: int = 10,
    spec: str = 'iso1',
    optimize=False,
    custom_spec: StlFormula | None = None,
    delta=0.1,
) -> StlMilpProblem:
    prob = create_stl_milp_problem(
        milp,
        N=N,
        delta=delta,
        gamma_0=0.0,
        gamma_N=10.0,
        gamma_unit_length=0.005,
        use_binary_expansion=True,
        n_gamma_digits=10,  # 0.005 * 2^10 = 5.12
    )

    # `iso1a` represents a modified version of `iso1` with different parameter set
    if spec[-1] == 'a':
        parameter_set = 'a'
        scenario_id = int(spec[-2])
    else:
        parameter_set = 'normal'
        scenario_id = int(spec[-1])

    car = prob.create_system_model()

    ax = car.add_state('ax', 0, 1000)
    bx = car.add_state('bx', 0, 1000)
    ax_vel = car.add_state('ax_vel', 10, 25)
    bx_vel = car.add_state('bx_vel', 10, 25)
    ax_acc = car.add_state('ax_acc', -10, 5)
    bx_acc = car.add_state('bx_acc', -10, 5)
    ay = car.add_state('ay', 0, 10.5)
    by = car.add_state('by', 0, 10.5)
    ay_vel = car.add_state('ay_vel', -10, 10)
    by_vel = car.add_state('by_vel', -10, 10)
    ay_acc = car.add_state('ay_acc', -3, 3)
    by_acc = car.add_state('by_acc', -3, 3)
    rssDistance_a = car.add_state('rssDistance_a', 0, 1000)
    rssDistance_b = car.add_state('rssDistance_b', 0, 1000)
    lateralRssDistance_a = car.add_state('lateralRssDistance_a', 0, 1000)
    lateralRssDistance_b = car.add_state('lateralRssDistance_b', 0, 1000)

    car.set_initial_state(
        ax=50.0,
        bx=(0, 100),
    )

    # Dynamics
    car.add_dynamics('ax_acc', -3, 1, constant=True)
    car.add_dynamics('ay_acc', -1, 1, constant=True)
    car.add_dynamics('bx_acc', -3, 1, constant=True)
    car.add_dynamics('by_acc', -1, 1, constant=True)

    car.add_double_integrator_dynamics('ax', 'ax_vel', 'ax_acc')
    car.add_double_integrator_dynamics('ay', 'ay_vel', 'ay_acc')
    car.add_double_integrator_dynamics('bx', 'bx_vel', 'bx_acc')
    car.add_double_integrator_dynamics('by', 'by_vel', 'by_acc')

    # extra constraints to define rss distances
    from stltspref import rss_distance as rss

    rss.setRssDistance(milp, rssDistance_a, rssDistance_b, ax_vel, bx_vel)
    rss.setLateralRssDistance(
        milp, lateralRssDistance_a, lateralRssDistance_b, ay_vel, by_vel
    )

    from stltspref.rss_scenario import car_length, car_width, get_danger, get_scenario

    def no_collision(margin=0.0):
        overlap_x = And(
            Atomic(L.f(1.0, 'ax', -1.0, 'bx') <= car_length + margin),
            Atomic(L.f(1.0, 'bx', -1.0, 'ax') <= car_length + margin),
        )
        overlap_y = And(
            Atomic(L.f(1.0, 'ay', -1.0, 'by') <= car_width + margin),
            Atomic(L.f(1.0, 'by', -1.0, 'ay') <= car_width + margin),
        )
        return Alw(And(overlap_x, overlap_y).negation())

    def no_crossing(car: str, margin=0.1):
        return Or(
            Atomic(L.unit(f'{car}y') <= 3.5 - margin),
            And(
                Atomic(L.unit(f'{car}y') <= 7.0 - margin),
                Atomic(L.unit(f'{car}y') >= 3.5 + car_width + margin),
            ),
            Atomic(L.unit(f'{car}y') >= 7.0 + car_width + margin),
        )

    ensure_monotonicity_by_derivative(milp, ax_vel.tolist(), prob.N)
    ensure_monotonicity_by_derivative(milp, bx_vel.tolist(), prob.N)
    ensure_monotonicity_by_derivative(milp, ay_vel.tolist(), prob.N)
    ensure_monotonicity_by_derivative(milp, by_vel.tolist(), prob.N)

    if optimize:
        alpha = milp.addMVar(1, lb=0.0, ub=20.0, name='alpha').item()
        milp.setObjective(alpha, gp.GRB.MAXIMIZE)
    else:
        alpha = 0.1

    if parameter_set == 'normal':
        safe_distance_margin = 1.0
    elif parameter_set == 'a':
        safe_distance_margin = 5.0
    else:
        raise ValueError(f'Unknown parameter set: {parameter_set}')
    if custom_spec is None:
        stl_spec = make_unique(
            And(
                get_danger(margin=1.0).negation(),
                get_scenario(scenario_id, alpha=alpha),
                no_collision(margin=safe_distance_margin),
                no_crossing('a'),
                no_crossing('b'),
            )
        )
    else:
        stl_spec = custom_spec
    prob.initialize_milp_formulation(stl_spec)

    return prob



from __future__ import annotations

import numpy as np
import pandas as pd

from stltspref.stl_base import StlFormula

from stltspref.problem import StlMilpProblem
from stltspref.validator import Approx


def get_robust_semantics_result(
    prob: StlMilpProblem, depth: int | None = None, stl_spec: StlFormula | None = None
) -> pd.DataFrame:
    if stl_spec is None:
        stl_spec = prob.stl_spec
    trace = prob.get_trace_result(interpolation=True)

    memo = {}
    for phi in stl_spec.get_subformulas(depth=depth):
        rob = phi.robust_semantics(trace, memo)
        memo[phi.name] = rob

    df = pd.DataFrame(
        memo,
        index=trace.time,
    )
    df.index.name = 't'
    return df


def validate_robust_semantics(prob: StlMilpProblem, depth=None) -> bool:
    trace = prob.get_trace_result(interpolation=True)
    stl_spec = prob.stl_spec
    gamma = prob.get_gamma_result()
    delta = prob.delta
    stl_sat = prob.milp_variables.stl_sat

    memo = {}
    for phi in stl_spec.get_subformulas(depth=depth):
        robustness = phi.robust_semantics(trace, memo)
        memo[phi.name] = robustness

        for k, Gamma in enumerate(zip(gamma[:-1], gamma[1:]), 1):
            target = (
                np.searchsorted(trace.time, Gamma[0]),
                np.searchsorted(trace.time, Gamma[1]),
            )
            if target[0] == target[1]:
                continue  # skip empty interval

            if stl_sat[phi][k].X >= 0.5:
                assert robustness[target[0] : target[1]].min() >= Approx(0), f'min robustness of `{phi.name}` in Gamma_{k}=[{Gamma[0]}, {Gamma[1]}] was {robustness[target[0]:target[1]].min()}'  # fmt: skip
            else:
                assert robustness[target[0] : target[1]].max() <= Approx(delta), f'max robustness of `{phi.name}` in Gamma_{k}=[{Gamma[0]}, {Gamma[1]}] was {robustness[target[0]:target[1]].max()}'  # fmt: skip

    return True



from __future__ import annotations

from .stl import (
    Alw,
    And,
    Atomic,
    Bottom,
    BoundedAlw,
    BoundedEv,
    Ev,
    Or,
    Release,
    StlFormula,
    Top,
    Until,
)


def str_interval(interval: tuple[float, float]) -> str:
    return f'[{interval[0]}, {interval[1]}]'


def str_atomic(
    variables,
    coefficients,
    constant,
    sense,
) -> str:
    combined = ' + '.join(
        f"{coeff:.5} * {var}" for var, coeff in zip(variables, coefficients)
    )
    if coefficients == [1] and constant == 0:
        return f'{variables[0]} {sense} 0'
    elif coefficients == [1]:
        return f'{variables[0]} {sense} {-constant:.5}'
    elif constant == 0:
        return f'{combined} {sense} 0'
    else:
        return f"{combined} + {constant:.5} {sense} 0"


def to_latex(phi: StlFormula) -> str:
    if isinstance(phi, Atomic):
        sanitized_name = str_atomic(
            variables=phi.predicate.inequality.expr.variables,
            coefficients=phi.predicate.inequality.expr.coefficients,
            constant=phi.predicate.inequality.expr.constant,
            sense=phi.predicate.inequality.sense,
        ).replace('_', '\\_')
        return f'\\text{{{sanitized_name}}}'
    elif isinstance(phi, Top):
        return '\\top'
    elif isinstance(phi, Bottom):
        return '\\bot'
    elif isinstance(phi, And):
        content = '\\wedge'.join([f'{to_latex(child)}' for child in phi.children])
        return f'({content})'
    elif isinstance(phi, Or):
        content = '\\vee'.join([f'{to_latex(child)}' for child in phi.children])
        return f'({content})'
    elif isinstance(phi, Alw):
        return f'\\Box {to_latex(phi.children[0])}'
    elif isinstance(phi, Ev):
        return f'\\Diamond {to_latex(phi.children[0])}'
    elif isinstance(phi, Until):
        return (
            f'({to_latex(phi.children[0])} \\mathop{{\\mathcal{{U}}}}'
            f' {to_latex(phi.children[1])})'
        )
    elif isinstance(phi, Release):
        return (
            f'({to_latex(phi.children[0])} \\mathop{{\\mathcal{{R}}}}'
            f' {to_latex(phi.children[1])})'
        )
    elif isinstance(phi, BoundedAlw):
        return f'\\Box_{{{str_interval(phi.interval)}}} {to_latex(phi.child)}'
    elif isinstance(phi, BoundedEv):
        return f'\\Diamond_{{{str_interval(phi.interval)}}} {to_latex(phi.child)}'
    else:
        raise ValueError(f"Unknown STL formula type: {type(phi)}")


def to_breach(phi: StlFormula) -> str:
    if isinstance(phi, Atomic):
        return str_atomic(
            variables=[v + '[t]' for v in phi.predicate.inequality.expr.variables],
            coefficients=phi.predicate.inequality.expr.coefficients,
            constant=phi.predicate.inequality.expr.constant,
            sense=phi.predicate.inequality.sense + '=',
        )
    elif isinstance(phi, Top):
        return 'true'
    elif isinstance(phi, Bottom):
        return 'false'
    elif isinstance(phi, And):
        return f'({" and ".join([to_breach(child) for child in phi.children])})'
    elif isinstance(phi, Or):
        return f'({" or ".join([to_breach(child) for child in phi.children])})'
    elif isinstance(phi, Alw):
        return f'(alw({to_breach(phi.children[0])}))'
    elif isinstance(phi, Ev):
        return f'(ev({to_breach(phi.children[0])}))'
    elif isinstance(phi, Until):
        return f'({to_breach(phi.children[0])} until {to_breach(phi.children[1])})'
    elif isinstance(phi, Release):
        return f'({to_breach(phi.children[0])} release {to_breach(phi.children[1])})'
    elif isinstance(phi, BoundedAlw):
        return f'(alw_{str_interval(phi.interval)} ({to_breach(phi.child)}))'
    elif isinstance(phi, BoundedEv):
        return f'(ev_{str_interval(phi.interval)} ({to_breach(phi.child)}))'
    else:
        raise ValueError(f"Unknown STL formula type: {type(phi)}")


def to_stlmc(phi: StlFormula) -> str:
    if isinstance(phi, Atomic):
        return str_atomic(
            variables=phi.predicate.inequality.expr.variables,
            coefficients=phi.predicate.inequality.expr.coefficients,
            constant=phi.predicate.inequality.expr.constant,
            sense=phi.predicate.inequality.sense + '=',
        )
    elif isinstance(phi, Top):
        return 'true'
    elif isinstance(phi, Bottom):
        return 'false'
    elif isinstance(phi, And):
        return f'({" and ".join([to_stlmc(child) for child in phi.children])})'
    elif isinstance(phi, Or):
        return f'({" or ".join([to_stlmc(child) for child in phi.children])})'
    elif isinstance(phi, Alw):
        return f'([][0, inf)({to_stlmc(phi.children[0])}))'
    elif isinstance(phi, Ev):
        return f'(<>[0, inf)({to_stlmc(phi.children[0])}))'
    elif isinstance(phi, Until):
        return f'({to_breach(phi.children[0])} U[0,inf) {to_stlmc(phi.children[1])})'
    elif isinstance(phi, Release):
        return f'({to_breach(phi.children[0])} R[0,inf) {to_stlmc(phi.children[1])})'
    elif isinstance(phi, BoundedAlw):
        return f'([]{str_interval(phi.interval)} ({to_stlmc(phi.child)}))'
    elif isinstance(phi, BoundedEv):
        return f'(<>{str_interval(phi.interval)} ({to_stlmc(phi.child)}))'
    else:
        raise ValueError(f"Unknown STL formula type: {type(phi)}")



import gurobipy as gp
from stltspref.preferential_synthesis import benchmark_pref_synth, diversity_finder
from stltspref.linear_expression import LinearExpression as L
from stltspref.problem import create_stl_milp_problem
from stltspref.stl import (
    Alw,
    And,
    Atomic,
    BoundedAlw,
    BoundedEv,
    BoundedUntil,
    Ev,
    Implies,
    Or,
    Until,
    StlFormula,
    make_unique,
)


# ---- PARAMETERS ----

benchmark_name = "rnc1"
diversity_mode = "random"
N = 10
numsols = 5
time_limit = 10000
gap = 0.01
absolute_gap = 3
export_traces = False
export_path = ""

# ---- EXAMPLE USING BENCHMARK ----

# benchmark_pref_synth(benchmark_name, diversity_mode, N, numsols, 
#                     time_limit=time_limit, gap=gap, absolute_gap=absolute_gap, 
#                     export_path=export_path, export_traces=export_traces)

# ---- EXAMPLE USING CUSTOM PROBLEM ----

milp = gp.Model()
prob = create_stl_milp_problem(
        milp,
        N=N,
        delta=0.1,
        gamma_N=20.0,
        gamma_unit_length=0.005,
        use_binary_expansion=True,
    )

elevator = prob.create_system_model()

elevator.add_state('z', 0, 40)
elevator.add_state('v', -5, 5)
elevator.add_state('a', -3, 3)

elevator.set_initial_state(z=(0,0), v=(0,0), a=(0,0))
elevator.add_dynamics('a', -3, 3, constant=True)
elevator.add_double_integrator_dynamics('z', 'v', 'a')

spec = make_unique(
    Ev(And(
        Atomic(L.unit('z') >= 20), 
        Ev(And(
            Atomic(L.unit('z') <= 1),
            Atomic(L.unit('v') >= 0)
            ))
    ))
)
prob.initialize_milp_formulation(spec)

diversity_finder(prob, diversity_mode, N, numsols, time_limit, 
                 gap, absolute_gap, export_traces, export_path)
